import logging
from math import ceil
from pathlib import Path

import dask.dataframe as dd
import numpy as np
import zarr

from ._constants import ExplorerConstants, FileNames
from .utils import explorer_file_path

log = logging.getLogger(__name__)


def subsample_indices(n_samples, factor: int = 4):
    n_sub = n_samples // factor
    return np.random.choice(n_samples, n_sub, replace=False)


def write_transcripts(
    path: Path,
    df: dd.DataFrame,
    gene: str = "gene",
    max_levels: int = 15,
    is_dir: bool = True,
    pixel_size: float = 0.2125,
):
    """Write a `transcripts.zarr.zip` file containing pyramidal transcript locations

    Args:
        path: Path to the Xenium Explorer directory where the transcript file will be written
        df: DataFrame representing the transcripts, with `"x"`, `"y"` column required, as well as the `gene` column (see the corresponding argument)
        gene: Column of `df` containing the genes names.
        max_levels: Maximum number of levels in the pyramid.
        is_dir: If `False`, then `path` is a path to a single file, not to the Xenium Explorer directory.
        pixel_size: Number of microns in a pixel. Invalid value can lead to inconsistent scales in the Explorer.
    """
    path = explorer_file_path(path, FileNames.POINTS, is_dir)

    # TODO: make everything using dask instead of pandas
    df = df.compute()

    num_transcripts = len(df)
    grid_size = ExplorerConstants.GRID_SIZE / ExplorerConstants.PIXELS_TO_MICRONS * pixel_size
    df[gene] = df[gene].astype("category")

    location = df[["x", "y"]]
    location *= pixel_size
    location = np.concatenate([location, np.zeros((num_transcripts, 1))], axis=1)

    if location.min() < 0:
        log.warning("Some transcripts are located outside of the image (pixels < 0)")
    log.info(f"Writing {len(df)} transcripts")

    xmax, ymax = location[:, :2].max(axis=0)

    gene_names = list(df[gene].cat.categories)
    num_genes = len(gene_names)

    codeword_gene_mapping = list(range(num_genes))

    valid = np.ones((num_transcripts, 1))
    uuid = np.stack([np.arange(num_transcripts), np.full(num_transcripts, 65535)], axis=1)
    transcript_id = np.stack([np.arange(num_transcripts), np.full(num_transcripts, 65535)], axis=1)
    gene_identity = df[gene].cat.codes.values[:, None]
    codeword_identity = np.stack([gene_identity[:, 0], np.full(num_transcripts, 65535)], axis=1)
    status = np.zeros((num_transcripts, 1))
    quality_score = np.full((num_transcripts, 1), ExplorerConstants.QUALITY_SCORE)

    ATTRS = {
        "codeword_count": num_genes,
        "codeword_gene_mapping": codeword_gene_mapping,
        "codeword_gene_names": gene_names,
        "gene_names": gene_names,
        "gene_index_map": dict(zip(gene_names, codeword_gene_mapping)),
        "number_genes": num_genes,
        "spatial_units": "micron",
        "coordinate_space": "refined-final_global_micron",
        "major_version": 4,
        "minor_version": 1,
        "name": "RnaDataset",
        "number_rnas": num_transcripts,
        "dataset_uuid": "unique-id-test",
        "data_format": 0,
    }

    GRIDS_ATTRS = {
        "grid_key_names": ["grid_x_loc", "grid_y_loc"],
        "grid_zip": False,
        "grid_size": [grid_size],
        "grid_array_shapes": [],
        "grid_number_objects": [],
        "grid_keys": [],
    }

    with zarr.ZipStore(path, mode="w") as store:
        g = zarr.group(store=store)
        g.attrs.put(ATTRS)

        grids = g.create_group("grids")

        for level in range(max_levels):
            log.info(f"   > Level {level}: {len(location)} transcripts")
            level_group = grids.create_group(level)

            tile_size = grid_size * 2**level

            indices = np.floor(location[:, :2] / tile_size).clip(0).astype(int)
            tiles_str_indices = np.array([f"{tx},{ty}" for (tx, ty) in indices])

            GRIDS_ATTRS["grid_array_shapes"].append([])
            GRIDS_ATTRS["grid_number_objects"].append([])
            GRIDS_ATTRS["grid_keys"].append([])

            n_tiles_x, n_tiles_y = ceil(xmax / tile_size), ceil(ymax / tile_size)

            for tx in range(n_tiles_x):
                for ty in range(n_tiles_y):
                    str_index = f"{tx},{ty}"
                    loc = np.where(tiles_str_indices == str_index)[0]

                    n_points_tile = len(loc)
                    chunks = (n_points_tile, 1)

                    if n_points_tile == 0:
                        continue

                    GRIDS_ATTRS["grid_array_shapes"][-1].append({})
                    GRIDS_ATTRS["grid_keys"][-1].append(str_index)
                    GRIDS_ATTRS["grid_number_objects"][-1].append(n_points_tile)

                    tile_group = level_group.create_group(str_index)
                    tile_group.array(
                        "valid",
                        valid[loc],
                        dtype="uint8",
                        chunks=chunks,
                    )
                    tile_group.array(
                        "status",
                        status[loc],
                        dtype="uint8",
                        chunks=chunks,
                    )
                    tile_group.array(
                        "location",
                        location[loc],
                        dtype="float32",
                        chunks=chunks,
                    )
                    tile_group.array(
                        "gene_identity",
                        gene_identity[loc],
                        dtype="uint16",
                        chunks=chunks,
                    )
                    tile_group.array(
                        "quality_score",
                        quality_score[loc],
                        dtype="float32",
                        chunks=chunks,
                    )
                    tile_group.array(
                        "codeword_identity",
                        codeword_identity[loc],
                        dtype="uint16",
                        chunks=chunks,
                    )
                    tile_group.array(
                        "uuid",
                        uuid[loc],
                        dtype="uint32",
                        chunks=chunks,
                    )
                    tile_group.array(
                        "id",
                        transcript_id[loc],
                        dtype="uint32",
                        chunks=chunks,
                    )

            if n_tiles_x * n_tiles_y == 1:
                GRIDS_ATTRS["number_levels"] = level + 1
                break

            sub_indices = subsample_indices(len(location))

            location = location[sub_indices]
            valid = valid[sub_indices]
            status = status[sub_indices]
            gene_identity = gene_identity[sub_indices]
            quality_score = quality_score[sub_indices]
            codeword_identity = codeword_identity[sub_indices]
            uuid = uuid[sub_indices]
            transcript_id = transcript_id[sub_indices]

        grids.attrs.put(GRIDS_ATTRS)



from .resolve import solve_conflicts, combine
from .stainings import StainingSegmentation
from ._tissue import tissue, shapes_bounding_box
from .methods import cellpose, baysor, custom_staining_based, comseg, stardist, proseg
from . import methods



import logging

import geopandas as gpd
import shapely
from shapely.geometry import GeometryCollection, MultiPolygon, Polygon

log = logging.getLogger(__name__)


def to_valid_polygons(geo_df: gpd.GeoDataFrame, simple_polygon: bool = True) -> gpd.GeoDataFrame:
    geo_df.geometry = geo_df.geometry.map(lambda cell: ensure_polygon(cell, simple_polygon))
    return geo_df[~geo_df.is_empty]


def ensure_polygon(
    cell: Polygon | MultiPolygon | GeometryCollection, simple_polygon: bool = True
) -> Polygon | MultiPolygon:
    """Ensures that the provided cell becomes a Polygon

    Args:
        cell: A shapely Polygon or MultiPolygon or GeometryCollection
        simple_polygon: If True, will return a Polygon without holes. Else, allow holes and MultiPolygon.

    Returns:
        The shape as a Polygon, or an empty Polygon if the cell was invalid
    """
    cell = shapely.make_valid(cell)

    if isinstance(cell, Polygon):
        if simple_polygon and cell.interiors:
            cell = Polygon(cell.exterior)
        return cell

    if isinstance(cell, MultiPolygon):
        return max(cell.geoms, key=lambda polygon: polygon.area) if simple_polygon else cell

    if isinstance(cell, GeometryCollection):
        geoms = [geom for geom in cell.geoms if isinstance(geom, Polygon)]

        if len(geoms) > 1 and not simple_polygon:
            return MultiPolygon(geoms)

        if geoms:
            return max(geoms, key=lambda polygon: polygon.area)

        log.warning(f"Found a cell of type {type(cell)} which contains no Polygon geometry")
        return Polygon()

    log.warning(f"Found a cell of unknown type {type(cell)}")
    return Polygon()



import typer

from .utils import SDATA_HELPER

app_resolve = typer.Typer()


@app_resolve.command()
def cellpose(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    cache_dir_name: list[str] = typer.Option(
        [],
        help="Name of the directories containing the cellpose segmentation on patches (or multiple directories if using multi-step segmentation). By default, uses the `cellpose_boundaries` directory",
    ),
):
    """Resolve patches conflicts after cellpose segmentation"""
    from sopa.constants import SopaKeys

    from .utils import _default_boundary_dir

    if not len(cache_dir_name):
        cache_dir_name = [SopaKeys.CELLPOSE_BOUNDARIES]

    patch_dir = [_default_boundary_dir(sdata_path, name) for name in cache_dir_name]

    _resolve_generic(sdata_path, patch_dir, SopaKeys.CELLPOSE_BOUNDARIES)


@app_resolve.command()
def stardist(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    cache_dir_name: list[str] = typer.Option(
        [],
        help="Name of the directories containing the stardist segmentation on patches (or multiple directories if using multi-step segmentation). By default, uses the `stardist_boundaries` directory",
    ),
):
    """Resolve patches conflicts after stardist segmentation"""
    from sopa.constants import SopaKeys

    from .utils import _default_boundary_dir

    if not len(cache_dir_name):
        cache_dir_name = [SopaKeys.STARDIST_BOUNDARIES]

    patch_dir = [_default_boundary_dir(sdata_path, name) for name in cache_dir_name]

    _resolve_generic(sdata_path, patch_dir, SopaKeys.STARDIST_BOUNDARIES)


@app_resolve.command()
def generic(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    method_name: str = typer.Option(
        help="Name of the method used during segmentation. This is also the key correspnding to the boundaries in `sdata.shapes`"
    ),
    cache_dir_name: list[str] = typer.Option(
        [],
        help="Name of the directories containing the generic segmentation on patches (or multiple directories if using multi-step segmentation). By default, uses the `<method_name>` directory",
    ),
):
    """Resolve patches conflicts after generic segmentation"""
    from .utils import _default_boundary_dir

    if not len(cache_dir_name):
        cache_dir_name = [method_name]

    patch_dir = [_default_boundary_dir(sdata_path, name) for name in cache_dir_name]

    _resolve_generic(sdata_path, patch_dir, method_name)


def _resolve_generic(sdata_path: str, patch_dirs: list[str], key_added: str):
    from sopa.io.standardize import read_zarr_standardized
    from sopa.segmentation import StainingSegmentation, solve_conflicts
    from sopa.utils import get_spatial_image

    sdata = read_zarr_standardized(sdata_path)

    assert len(patch_dirs) > 0, "No patch directory was provided, cannot load cells"

    image_key, _ = get_spatial_image(sdata, return_key=True)

    cells = StainingSegmentation.read_patches_cells(patch_dirs)
    cells = solve_conflicts(cells)

    StainingSegmentation.add_shapes(sdata, cells, image_key, key_added)


@app_resolve.command()
def baysor(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    gene_column: str = typer.Option(None, help="Column of the transcripts dataframe containing the genes names"),
    min_area: float = typer.Option(0, help="Cells with an area less than this value (in microns^2) will be filtered"),
):
    """Resolve patches conflicts after baysor segmentation."""
    import sopa
    from sopa.constants import SopaKeys
    from sopa.io.standardize import read_zarr_standardized
    from sopa.segmentation.methods.utils import resolve

    sdata = read_zarr_standardized(sdata_path)

    patches_dirs = sopa.utils.get_transcripts_patches_dirs(sdata)

    resolve(
        sdata,
        patches_dirs,
        gene_column,
        min_area=min_area,
        key_added=SopaKeys.BAYSOR_BOUNDARIES,
    )


@app_resolve.command()
def comseg(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    gene_column: str = typer.Option(None, help="Column of the transcripts dataframe containing the genes names"),
    min_area: float = typer.Option(0, help="Cells with an area less than this value (in microns^2) will be filtered"),
):
    """Resolve patches conflicts after comseg segmentation."""
    import sopa
    from sopa.constants import SopaKeys
    from sopa.io.standardize import read_zarr_standardized
    from sopa.segmentation.methods.utils import resolve

    sdata = read_zarr_standardized(sdata_path)

    patches_dirs = sopa.utils.get_transcripts_patches_dirs(sdata)

    resolve(
        sdata,
        patches_dirs,
        gene_column,
        min_area=min_area,
        key_added=SopaKeys.COMSEG_BOUNDARIES,
    )



import torch
from torch import nn


class DINOv2Features(nn.Module):
    def __init__(self, model="dinov2_vitl14"):
        super().__init__()
        self.dinov2_vitl14 = torch.hub.load("facebookresearch/dinov2", model)
        self.dinov2_vitl14.eval()

    def forward(self, x):
        feats = self.dinov2_vitl14.forward_features(x)
        return feats["x_norm_clstoken"]



import numpy as np
from anndata import AnnData
from scipy.spatial import Delaunay
from shapely.geometry import LineString, Polygon


class Component:
    def __init__(self, adata: AnnData, delaunay: Delaunay, neighbors: np.ndarray) -> None:
        """Visitis one niche component to find border edges. Then, it assembles these border edges
        together to create the rings used as input to `shapely.geometry.Polygon`

        Args:
            adata: An `AnnData` object
            delaunay: The corresponding Delaunay graph
            neighbors: The neighbors of the delaunay graph
        """
        self.adata = adata
        self.delaunay = delaunay
        self.neighbors = neighbors

        self.lines: dict[int, list[int]] = {}  # values are ordered vertices indices
        self.limits: dict[int, tuple[int]] = {}  # values are (line_index, 0 if left / 1 if right)
        self._counter = 0

    def first_vertex_index(self) -> int:
        return next(iter(self.lines.values()))[0]

    def _standard_add(self, i: int, j: int, line_index_i: int, right_i: int):
        if right_i:
            del self.limits[self.lines[line_index_i][-1]]
            self.lines[line_index_i].append(j)
            self.limits[j] = (line_index_i, -1)
        else:
            del self.limits[i]
            self.lines[line_index_i].insert(0, j)
            self.limits[j] = (line_index_i, 0)

    def _insert_new(self, i: int, j: int):
        self._counter += 1
        self.lines[self._counter] = [i, j]
        self.limits[i] = (self._counter, 0)
        self.limits[j] = (self._counter, -1)

    def add_edge(self, i: int, j: int):
        line_index_i, right_i = self.limits.get(i, (None, None))
        line_index_j, right_j = self.limits.get(j, (None, None))

        if not line_index_i and not line_index_j:
            self._insert_new(i, j)

        elif line_index_i and not line_index_j:
            self._standard_add(i, j, line_index_i, right_i)

        elif line_index_j and not line_index_i:
            self._standard_add(j, i, line_index_j, right_j)

        else:  # join both lines
            del self.limits[j]
            del self.limits[i]

            if line_index_i == line_index_j:  # cycle
                self.lines[line_index_i].append(i if right_j else j)
                return

            line_i, line_j = self.lines[line_index_i], self.lines[line_index_j]
            del self.lines[line_index_j]
            self.limits[line_j[0 if right_j else -1]] = [line_index_i, right_i]

            if not (right_i ^ right_j):
                line_j = line_j[::-1]

            self.lines[line_index_i] = line_i + line_j if right_i else line_j + line_i

    def to_line(self, indices: list[int]) -> LineString:
        return LineString(self.adata.obsm["spatial"][indices])

    def __len__(self) -> int:
        return len(self.lines)

    @property
    def rings(self) -> list[LineString]:
        return [self.to_line(indices) for indices in self.lines.values()]

    @property
    def polygon(self) -> Polygon:
        rings = self.rings
        index_largest = np.argmax([Polygon(ring).area for ring in rings])

        return Polygon(rings[index_largest], rings[:index_largest] + rings[index_largest + 1 :]).buffer(0)

    def visit(self, simplices_to_visit: set[int]):
        visited = set()
        component_simplices_to_visit = {simplices_to_visit.pop()}

        while component_simplices_to_visit:
            simplex_index = component_simplices_to_visit.pop()

            simplices_to_visit.discard(simplex_index)
            visited.add(simplex_index)

            simplex = self.delaunay.simplices[simplex_index]
            ngh_simplex_indices = self.neighbors[simplex_index]

            for i in range(3):
                ngh_simplex_index = ngh_simplex_indices[i]
                if ngh_simplex_index == -1:  # 'simplex' is at the boundary
                    self.add_edge(simplex[(i + 1) % 3], simplex[(i + 2) % 3])
                elif ngh_simplex_index not in visited:
                    component_simplices_to_visit.add(ngh_simplex_index)



import logging
from pathlib import Path

import numpy as np
from spatialdata import SpatialData
from spatialdata.transformations import Scale, set_transformation

from ...constants import SopaAttrs
from ...segmentation import shapes_bounding_box
from ...utils import ensure_string_channel_names, get_spatial_image
from .utils import _default_image_kwargs

log = logging.getLogger(__name__)


def visium_hd(
    path: str | Path,
    fullres_image_file: str | Path | None = None,
    dataset_id: str | None = None,
    image_models_kwargs: dict | None = None,
    imread_kwargs: dict | None = None,
    var_names_make_unique: bool = True,
    bins_as_squares: bool = True,
    **kwargs: int,
) -> SpatialData:
    """Read Visium HD data as a `SpatialData` object. For more information, refer to [spatialdata-io](https://spatialdata.scverse.org/projects/io/en/latest/generated/spatialdata_io.visium_hd.html).

    !!! info
        If your `fullres_image_file` is not in the `microscope_image` directory, you can specify the path to the image file using the `fullres_image_file` argument.

    Args:
        path: Path to the Visium HD directory containing all the experiment files
        fullres_image_file: Path to the full-resolution image. By default the image is searched in the `'microscope_image'` directory.
        dataset_id: Identifier of the dataset. By default, inferred from the prefix of the input files. If the files have no prefix (e.g., `feature_slice.h5`), use `dataset_id=""`.
        image_models_kwargs: Keyword arguments passed to `spatialdata.models.Image2DModel`.
        imread_kwargs: Keyword arguments passed to `dask_image.imread.imread`.
        var_names_make_unique: If True, ensure that the var names are unique.
        bins_as_squares: If `True`, the bins are represented as squares. If `False`, the bins are represented as circles. For a correct visualization one should use squares.
        kwargs: Additional keyword arguments passed to `spatialdata_io.visium_hd`.

    Returns:
        A `SpatialData` object representing the Xenium experiment
    """
    from spatialdata_io.readers.visium_hd import visium_hd as visium_hd_spatialdata_io

    image_models_kwargs, imread_kwargs = _default_image_kwargs(image_models_kwargs, imread_kwargs)

    del image_models_kwargs["scale_factors"]  # already set in the spatialdata_io reader

    sdata: SpatialData = visium_hd_spatialdata_io(
        path,
        dataset_id=dataset_id,
        fullres_image_file=fullres_image_file,
        image_models_kwargs=image_models_kwargs,
        imread_kwargs=imread_kwargs,
        bins_as_squares=bins_as_squares,
        **kwargs,
    )

    ensure_string_channel_names(sdata)  # Ensure that channel names are strings

    ### Add Sopa attributes to detect the spatial elements
    for key in sdata.images:
        if key.endswith("_full_image"):
            sdata.attrs[SopaAttrs.CELL_SEGMENTATION] = key
        elif key.endswith("_hires_image"):
            sdata.attrs[SopaAttrs.TISSUE_SEGMENTATION] = key

    for key, adata in sdata.tables.items():
        if key.endswith("_002um"):
            sdata.attrs[SopaAttrs.BINS_TABLE] = key
        if var_names_make_unique:
            adata.var_names_make_unique()

    _set_microns_coordinate_system(sdata, bins_as_squares=bins_as_squares)

    for key in sdata.shapes:
        if key.endswith("_cell_segmentations"):
            sdata.attrs[SopaAttrs.BOUNDARIES] = key
            break

    for key in sdata.shapes:
        if key.endswith("_002um"):
            shapes_bounding_box(sdata, key)
            break

    _sanity_check_images(sdata)

    return sdata


def _set_microns_coordinate_system(sdata: SpatialData, bins_as_squares: bool = True) -> None:
    for key, geo_df in sdata.shapes.items():
        if not key[-5:-2].isdigit():
            continue

        diameter = int(key[-5:-2])

        area = geo_df.geometry.iloc[0].area
        microns_per_pixel = diameter / (np.sqrt(area) if bins_as_squares else geo_df["radius"].iloc[0] * 2)

        scale = Scale([microns_per_pixel, microns_per_pixel], axes=("x", "y"))
        set_transformation(geo_df, scale, "microns", set_all=False)


def _sanity_check_images(sdata: SpatialData) -> None:
    if SopaAttrs.CELL_SEGMENTATION not in sdata.attrs:
        if SopaAttrs.BOUNDARIES not in sdata.attrs:
            log.warning(
                "The full-resolution image was not found, you'll not be able to run cell segmentation.\nPlease set the `fullres_image_file` argument to the path of the full-resolution image."
            )
        else:
            log.info(
                "The full-resolution image was not found, but you have prior 10X Genomics segmentation. You can run `sopa.segmentation.proseg` with `prior_shapes_key='auto'`."
            )
        return

    if SopaAttrs.TISSUE_SEGMENTATION not in sdata.attrs:
        log.warning("The high-resolution image was not found, you'll not be able to run tissue segmentation.")
        return

    shape1 = get_spatial_image(sdata, sdata.attrs[SopaAttrs.CELL_SEGMENTATION]).shape
    shape2 = get_spatial_image(sdata, sdata.attrs[SopaAttrs.TISSUE_SEGMENTATION]).shape

    if (shape1[1] * shape1[2]) < 5_000**2 or (shape1[1] * shape1[2]) / (shape2[1] * shape2[2]) < 4:
        log.warning(
            f"The full-resolution image has a shape {shape1} which seems small. Ensure it is the full-resolution microscopy image (usually a file of 1GB+), not the CytAssist hires image."
        )



import logging

import anndata
import geopandas as gpd
import numpy as np
import pandas as pd
from anndata import AnnData
from shapely import Polygon
from spatialdata import SpatialData

from ..constants import SopaAttrs, SopaKeys
from ..utils import get_feature_key, get_spatial_element, to_intrinsic
from .aggregation import Aggregator

log = logging.getLogger(__name__)


def overlay_segmentation(
    sdata: SpatialData,
    shapes_key: str,
    gene_column: str | None = None,
    area_ratio_threshold: float = 0.25,
    image_key: str | None = None,
    table_key: str = SopaKeys.TABLE,
):
    """Overlay a segmentation on top of an existing segmentation

    Args:
        sdata: A `SpatialData` object
        shapes_key: The key of the new shapes to be added
        gene_column: Key of the points dataframe containing the genes names
        area_ratio_threshold: Threshold between 0 and 1. For each original cell overlapping with a new cell, we compute the overlap-area/cell-area, if above the threshold the cell is removed.
        image_key: Optional key of the original image
        table_key: Key of the table to be overlayed
    """
    aggregate_genes, aggregate_channels = False, False

    assert table_key in sdata.tables, f"No table with name '{table_key}' found in the SpatialData object"

    old_table: AnnData = sdata.tables[table_key]

    assert SopaKeys.UNS_KEY in old_table.uns, "It seems the table was not aggregated using `sopa.aggregate`"

    sopa_attrs = old_table.uns[SopaKeys.UNS_KEY]

    aggregate_genes = sopa_attrs[SopaKeys.UNS_HAS_TRANSCRIPTS]
    aggregate_channels = sopa_attrs[SopaKeys.UNS_HAS_INTENSITIES]

    if aggregate_genes and gene_column is None:
        points = get_spatial_element(sdata.points, key=sdata.attrs.get(SopaAttrs.TRANSCRIPTS))
        gene_column = get_feature_key(points, raise_error=True)

    aggregator = Aggregator(sdata, image_key=image_key, shapes_key=shapes_key)
    aggregator.sdata.tables[f"{SopaKeys.OLD_TABLE_PREFFIX}{table_key}"] = old_table
    del aggregator.sdata.tables[table_key]

    old_shapes_key = old_table.uns["spatialdata_attrs"]["region"]
    instance_key = old_table.uns["spatialdata_attrs"]["instance_key"]

    if isinstance(old_shapes_key, list):
        assert len(old_shapes_key) == 1, "Can't overlap segmentation on multi-region SpatialData object"
        old_shapes_key = old_shapes_key[0]

    old_geo_df = aggregator.sdata[old_shapes_key]
    geo_df = to_intrinsic(aggregator.sdata, aggregator.geo_df, old_geo_df)

    geo_df.index.name = None
    gdf_join = gpd.sjoin(old_geo_df, geo_df)
    gdf_join["geometry_right"] = gdf_join["index_right"].map(lambda i: geo_df.geometry.iloc[i])
    gdf_join["overlap_ratio"] = gdf_join.apply(_overlap_area_ratio, axis=1)
    gdf_join: gpd.GeoDataFrame = gdf_join[gdf_join.overlap_ratio >= area_ratio_threshold]

    table_crop = old_table[~np.isin(old_table.obs[instance_key], gdf_join.index)].copy()
    table_crop.obs[SopaKeys.CELL_OVERLAY_KEY] = False

    aggregator.compute_table(
        aggregate_channels=aggregate_channels,
        aggregate_genes=aggregate_genes,
        gene_column=gene_column,
        key_added=table_key,
    )
    aggregator.table.obs[SopaKeys.CELL_OVERLAY_KEY] = True

    aggregator.table = anndata.concat(
        [table_crop, aggregator.table],
        uns_merge="first",
        join="outer",
    )

    aggregator.shapes_key = f"{old_shapes_key}_overlay_{aggregator.shapes_key}"

    geo_df_cropped = old_geo_df.loc[~old_geo_df.index.isin(gdf_join.index)]
    aggregator.geo_df = pd.concat([geo_df_cropped, geo_df], join="outer", axis=0)
    aggregator.geo_df.attrs = old_geo_df.attrs

    aggregator.add_standardized_table(table_key)


def _overlap_area_ratio(row) -> float:
    poly: Polygon = row["geometry"]
    poly_right: Polygon = row["geometry_right"]
    return poly.intersection(poly_right).area / poly.area



import dask_image.ndinterp
import numpy as np
from spatialdata import SpatialData
from xarray import DataArray, DataTree

from . import get_spatial_image


def resize_numpy(arr: np.ndarray, scale_factor: float, dims: list[str], output_shape: list[int]) -> np.ndarray:
    """Resize a numpy image

    Args:
        arr: a `numpy` array
        scale_factor: Scale factor of resizing, e.g. `2` will decrease the width by 2
        dims: List of dimension names. Only `"x"` and `"y"` are resized.
        output_shape: Size of the output array

    Returns:
        Resized array
    """
    resize_dims = [dim in ["x", "y"] for dim in dims]
    transform = np.diag([scale_factor if resize_dim else 1 for resize_dim in resize_dims])

    return dask_image.ndinterp.affine_transform(arr, matrix=transform, output_shape=output_shape).compute()


def assert_is_integer_dtype(dtype: np.dtype):
    assert np.issubdtype(dtype, np.integer), f"Expecting image to have an integer dtype, but found {dtype}"


def scale_dtype(arr: np.ndarray, dtype: np.dtype) -> np.ndarray:
    """Change the dtype of an array but keep the scale compared to the type maximum value.

    !!! note "Example"
        For an array of dtype `uint8` being transformed to `np.uint16`, the value `255` will become `65535`

    Args:
        arr: A `numpy` array
        dtype: Target `numpy` data type

    Returns:
        A scaled `numpy` array with the dtype provided.
    """
    assert_is_integer_dtype(arr.dtype)
    assert_is_integer_dtype(dtype)

    if arr.dtype == dtype:
        return arr

    factor = np.iinfo(dtype).max / np.iinfo(arr.dtype).max
    return (arr * factor).astype(dtype)


def get_channel_names(image: DataArray | DataTree | SpatialData, image_key: str | None = None) -> np.ndarray:
    """Get the channel names of an image or a SpatialData object.

    Args:
        image: Either a `DataArray`, a `DataTree`, or a `SpatialData` object. If a `SpatialData` object, the `image_key` argument can be used.
        image_key: If `image` is a SpatialData object, the key of the image to get the channel names from. If `None`, tries to get it automatically.

    Returns:
        An array of channel names.
    """
    if isinstance(image, SpatialData):
        image = get_spatial_image(image, key=image_key)

    if isinstance(image, DataArray):
        return image.coords["c"].values
    if isinstance(image, DataTree):
        return image["scale0"].coords["c"].values
    raise ValueError(f"Image must be a DataTree or a DataArray. Found: {type(image)}")


def validated_channel_names(image: DataArray | DataTree | SpatialData, image_key: str | None = None) -> list[str]:
    channel_names: list[str] = list(get_channel_names(image, image_key=image_key))
    return [c.replace("/", "_") for c in channel_names]


def is_valid_c_coords(c_coords: np.ndarray) -> bool:
    return c_coords.dtype.kind in {"U", "S", "O"}


def ensure_string_channel_names(sdata: SpatialData, default_single_channel: str | None = "DAPI"):
    for key, image in list(sdata.images.items()):
        c_coords = get_channel_names(image)

        if is_valid_c_coords(c_coords):
            continue

        c_coords = [str(i) for i in range(len(c_coords))]
        if len(c_coords) == 1 and default_single_channel is not None:
            c_coords = [default_single_channel]

        sdata.set_channel_names(key, c_coords)



import logging
import re
from pathlib import Path

import dask.array as da
import geopandas as gpd
import numpy as np
import pandas as pd
import tifffile
import xarray as xr
from anndata import AnnData
from dask_image.imread import imread
from scipy.sparse import csr_matrix
from shapely.geometry import Polygon
from spatialdata import SpatialData
from spatialdata.models import Image2DModel, Labels2DModel, PointsModel, ShapesModel, TableModel
from spatialdata.transformations import Scale

from ...constants import SopaAttrs
from .utils import _deduplicate_names, _default_image_kwargs

log = logging.getLogger(__name__)

COSMX_PIXEL_SIZE = 0.120280945


def cosmx(
    path: str | Path,
    dataset_id: str | None = None,
    fov: int | None = None,
    read_image: bool = True,
    read_proteins: bool = False,
    cells_labels: bool = False,
    cells_table: bool = False,
    cells_polygons: bool = False,
    image_models_kwargs: dict | None = None,
    labels_models_kwargs: dict | None = None,
    imread_kwargs: dict | None = None,
    flip_image: bool = False,
    fov_shift: bool | None = None,
) -> SpatialData:
    """
    Read *CosMx Nanostring / Bruker* data. The fields of view are stitched together, except if `fov` is provided.

    This function reads the following files:
        - `*_fov_positions_file.csv` or `*_fov_positions_file.csv.gz`: FOV locations
        - `Morphology2D` directory: all the FOVs morphology images
        - `*_tx_file.csv.gz` or `*_tx_file.csv`: Transcripts location and names
        - If `read_proteins` is `True`, all the images under the nested `ProteinImages` directories will be read

        These files must be exported as flat files in AtomX. That is: within a study, click on "Export" and then select files from the "Flat CSV Files" section (transcripts flat and FOV position flat).

    Args:
        path: Path to the root directory containing *Nanostring* files.
        dataset_id: Optional name of the dataset (needs to be provided if not inferred).
        fov: Number of one single field of view to be read. If not provided, reads all FOVs and create a stitched image.
        read_image: Whether to read the images or not.
        read_proteins: Whether to read the proteins or the transcripts.
        cells_labels: Whether to read the cell labels or not.
        cells_table: Whether to read the cell table or not.
        cells_polygons: Whether to read the cell polygons or not.
        image_models_kwargs: Keyword arguments passed to `spatialdata.models.Image2DModel`.
        imread_kwargs: Keyword arguments passed to `dask_image.imread.imread`.
        flip_image: For some buggy exports of AtomX 1.3.2, `flip_image=True` has to be used for stitching. See [this](https://github.com/gustaveroussy/sopa/issues/231) issue.
        fov_shift: Whether to apply FOV shift correction. For some datasets, there is a one-FOV shift in the y direction between the image and the polygons/transcripts. If `None`, it will be inferred automatically based on the FOV positions file and the polygons files.

    Returns:
        A `SpatialData` object representing the CosMx experiment
    """
    path = Path(path)

    assert read_image or cells_labels, "At least one of `read_image` or `cells_labels` must be True."

    image_models_kwargs, imread_kwargs = _default_image_kwargs(image_models_kwargs, imread_kwargs)
    labels_models_kwargs = {"chunks": (4256, 4256)} if labels_models_kwargs is None else labels_models_kwargs

    dataset_id = _infer_dataset_id(path, dataset_id)

    _reader = _CosMXReader(path, dataset_id, fov, flip_image, fov_shift)
    attrs = {}

    ### Read elements
    region = _reader.shapes_key if cells_polygons else (_reader.labels_key if cells_labels else None)
    tables = _reader.read_tables(region) if cells_table else {}

    images = {}
    if read_image:
        images = _reader.read_images(
            read_proteins=read_proteins,
            imread_kwargs=imread_kwargs,
            image_models_kwargs=image_models_kwargs,
        )
        attrs[SopaAttrs.CELL_SEGMENTATION] = next(iter(images))

    labels = (
        _reader.read_labels(imread_kwargs=imread_kwargs, labels_models_kwargs=labels_models_kwargs)
        if cells_labels
        else {}
    )
    shapes = _reader.read_shapes() if cells_polygons else {}

    points = _reader.read_transcripts() if not read_proteins else {}
    if points:
        attrs[SopaAttrs.TRANSCRIPTS] = next(iter(points.keys()))
        attrs[SopaAttrs.PRIOR_TUPLE_KEY] = ("global_cell_id", 0)

    return SpatialData(images=images, labels=labels, points=points, tables=tables, shapes=shapes, attrs=attrs)


class _CosMXReader:
    def __init__(self, path: Path, dataset_id: str, fov: int | None, flip_image: bool, fov_shift: bool | None):
        self.path = path
        self.dataset_id = dataset_id
        self.fov = fov
        self.flip_image = flip_image
        self.fov_shift = fov_shift

        self.fov_locs = self._read_fov_locs()

        if self.fov is not None:
            log.info(f"Reading single FOV ({self.fov}), the image will not be stitched")
        elif fov_shift is None:
            self.fov_shift = self._infer_fov_shift()
            log.info(
                f"FOV shift correction is {'enabled' if self.fov_shift else 'disabled'} (if this is not correct, please set `fov_shift` manually)"
            )

    def read_transcripts(self):
        transcripts_file = self.path / f"{self.dataset_id}_tx_file.csv.gz"

        if transcripts_file.exists():
            df = pd.read_csv(transcripts_file, compression="gzip")
        else:
            transcripts_file = self.path / f"{self.dataset_id}_tx_file.csv"
            assert transcripts_file.exists(), f"Transcript file {transcripts_file} not found."
            df = pd.read_csv(transcripts_file)

        TRANSCRIPT_COLUMNS = ["x_global_px", "y_global_px", "target"]
        assert np.isin(TRANSCRIPT_COLUMNS, df.columns).all(), (
            f"The file {transcripts_file} must contain the following columns: {', '.join(TRANSCRIPT_COLUMNS)}. Consider using a different export module."
        )

        df["global_cell_id"] = self._get_global_cell_id(df)

        if self.fov is None:
            df["x"] = (df["x_global_px"] - self.fov_locs["xmin"].min()) * COSMX_PIXEL_SIZE
            df["y"] = (df["y_global_px"] - self.fov_locs["ymin"].min()) * COSMX_PIXEL_SIZE
            points_name = "points"
        else:
            df = df[df["fov"] == self.fov]
            df["x"] = df["x_local_px"] * COSMX_PIXEL_SIZE
            df["y"] = df["y_local_px"] * COSMX_PIXEL_SIZE
            points_name = f"F{self.fov:0>5}_points"

        from spatialdata_io._constants._constants import CosmxKeys

        transcripts = PointsModel.parse(
            df,
            feature_key=CosmxKeys.TARGET_OF_TRANSCRIPT,
            transformations={"global": Scale([1 / COSMX_PIXEL_SIZE, 1 / COSMX_PIXEL_SIZE], axes=("x", "y"))},
        )

        return {points_name: transcripts}

    def _read_cell_metadata(self) -> pd.DataFrame:
        from spatialdata_io._constants._constants import CosmxKeys

        metadata = self._read_csv_gz(f"{self.dataset_id}_{CosmxKeys.METADATA_SUFFIX}")
        metadata.index = self._get_global_cell_id(metadata)

        if "cell_id" in metadata.columns:
            del metadata["cell_id"]

        return metadata

    def read_tables(self, region: str | None) -> dict[str, AnnData]:
        from spatialdata_io._constants._constants import CosmxKeys

        counts = self._read_csv_gz(f"{self.dataset_id}_{CosmxKeys.COUNTS_SUFFIX}")
        counts = counts[counts["cell_ID"] != 0]  # remove background
        counts.index = self._get_global_cell_id(counts)
        counts.drop(columns=["fov", "cell_ID"], inplace=True)

        obs = self._read_cell_metadata()

        assert (obs.index == counts.index).all(), "The cell IDs in the metadata and counts files do not match."

        obs[CosmxKeys.FOV] = pd.Categorical(obs[CosmxKeys.FOV].astype(str))
        if region is not None:
            obs["region_key"] = pd.Series(region, index=obs.index, dtype="category")
            obs["global_cell_id"] = obs.index
        obs.index = obs.index.astype(str)

        adata = AnnData(csr_matrix(counts.values), obs=obs, var=pd.DataFrame(index=counts.columns))

        table = TableModel.parse(
            adata,
            region=region,
            region_key="region_key" if region is not None else None,
            instance_key="global_cell_id" if region is not None else None,
        )

        table.obsm["spatial"] = table.obs[[CosmxKeys.X_GLOBAL_CELL, CosmxKeys.Y_GLOBAL_CELL]].to_numpy()
        table.obs.drop(columns=[CosmxKeys.X_GLOBAL_CELL, CosmxKeys.Y_GLOBAL_CELL], inplace=True)

        return {"table": table}

    def _infer_fov_shift(self) -> bool:
        if self.flip_image:
            return True

        try:
            df_poly = self.df_poly
        except FileNotFoundError:
            log.warning("Polygons file not found, cannot infer FOV shift.")
            return False

        fov_y_min = df_poly.groupby("fov")["y_global_px"].min()

        return (fov_y_min < self.fov_locs.loc[fov_y_min.index, "y"]).mean() > 0.1

    @property
    def df_poly(self) -> pd.DataFrame:
        if not hasattr(self, "_df_poly"):
            self._df_poly = self._read_csv_gz(f"{self.dataset_id}-polygons.csv")
        return self._df_poly

    def read_shapes(self) -> dict[str, gpd.GeoDataFrame]:
        df_poly = self.df_poly
        df_poly.rename(columns={"cellID": "cell_ID"}, inplace=True)

        if self.flip_image:
            x_origin = self.fov_locs.loc[df_poly["fov"], "xmin"].values
            y_origin = self.fov_locs.loc[df_poly["fov"], "ymin"].values
            df_poly["x_global_px"] = x_origin + (x_origin - df_poly["x_global_px"])
            df_poly["y_global_px"] = y_origin + (y_origin - df_poly["y_global_px"])

        if self.fov is None:
            df_poly.index = self._get_global_cell_id(df_poly)
            xy_keys = ["x_global_px", "y_global_px"]
        else:
            df_poly = df_poly[df_poly["fov"] == self.fov]
            df_poly.index = df_poly["cell_ID"]
            xy_keys = ["x_local_px", "y_local_px"]

        geometry = df_poly.groupby(level=0).apply(lambda sub_df: Polygon(sub_df[xy_keys]))
        gdf = gpd.GeoDataFrame(df_poly.groupby(level=0)[["fov"]].first(), geometry=geometry)

        if self.fov is None:
            gdf.geometry = gdf.geometry.translate(-self.fov_locs["xmin"].min(), -self.fov_locs["ymin"].min())

        return {self.shapes_key: ShapesModel.parse(gdf)}

    @property
    def shapes_key(self) -> str:
        return f"F{self.fov:0>5}_cells_polygons" if self.fov is not None else "cells_polygons"

    @property
    def labels_key(self) -> str:
        return f"F{self.fov:0>5}_labels" if self.fov is not None else "stitched_labels"

    def read_images(
        self, read_proteins: bool, imread_kwargs: dict, image_models_kwargs: dict
    ) -> dict[str, xr.DataArray | xr.DataTree]:
        images_dir = _find_dir(self.path, "Morphology2D")
        morphology_coords = _get_morphology_coords(images_dir)

        protein_dir_dict = {}
        if read_proteins:
            protein_dir_dict = {
                int(protein_dir.parent.name[3:]): protein_dir
                for protein_dir in list(self.path.rglob("**/FOV*/ProteinImages"))
            }
            assert len(protein_dir_dict), f"No directory called 'ProteinImages' was found under {self.path}"

        if self.fov is None:
            return {
                "stitched_image": self._stitch_tifffiles(
                    images_dir,
                    protein_dir_dict=protein_dir_dict,
                    morphology_coords=morphology_coords,
                    imread_kwargs=imread_kwargs,
                    image_models_kwargs=image_models_kwargs,
                )
            }

        fov_file = _find_matching_fov_file(images_dir, self.fov)
        image, c_coords = _read_fov_image(fov_file, protein_dir_dict.get(self.fov), morphology_coords, **imread_kwargs)
        image = Image2DModel.parse(image, dims=("c", "y", "x"), c_coords=c_coords, **image_models_kwargs)

        return {f"F{self.fov:0>5}_image": image}

    def read_labels(self, imread_kwargs: dict, labels_models_kwargs: dict) -> dict[str, xr.DataArray | xr.DataTree]:
        labels_dir = _find_dir(self.path, "CellLabels")

        if self.fov is None:
            return {
                self.labels_key: self._stitch_tifffiles(
                    labels_dir,
                    labels=True,
                    imread_kwargs=imread_kwargs,
                    image_models_kwargs=labels_models_kwargs,
                )
            }

        fov_file = _find_matching_fov_file(labels_dir, self.fov)
        labels, _ = _read_fov_image(fov_file, None, [], **imread_kwargs)
        labels = Labels2DModel.parse(labels[0], dims=("y", "x"), **labels_models_kwargs)

        return {self.labels_key: labels}

    def _stitch_tifffiles(
        self,
        images_dir: Path,
        imread_kwargs: dict,
        image_models_kwargs: dict,
        protein_dir_dict: dict | None = None,
        morphology_coords: list[str] | None = None,
        labels: bool = False,
    ) -> xr.DataArray | xr.DataTree:
        images_paths = list(images_dir.glob("*.[Tt][Ii][Ff]"))
        shape = imread(images_paths[0]).shape[1:]

        self._set_fov_locs_bounding_boxes(shape)

        fov_images, c_coords_dict = {}, {}
        pattern = re.compile(r".*_F(\d+)")

        protein_dir_dict = protein_dir_dict or {}
        morphology_coords = morphology_coords or []

        if labels:
            obs = self._read_cell_metadata()

        for image_path in images_paths:
            fov = int(pattern.findall(image_path.name)[0])

            image, c_coords = _read_fov_image(image_path, protein_dir_dict.get(fov), morphology_coords, **imread_kwargs)
            assert image.shape[1:] == shape, (
                f"Expected all images to have the same shape {shape}, but found {image.shape[1:]} for FOV {fov}."
            )

            if labels and (max_label := image.max().compute()) > 0:
                fov_obs = obs[obs["fov"] == fov]
                local_ids, global_ids = fov_obs["cell_ID"], fov_obs.index

                mapping = np.zeros(max(max_label, local_ids.max()) + 1, dtype=int)
                mapping[local_ids] = global_ids

                image = da.map_blocks(mapping.__getitem__, image, dtype=int)

            fov_images[fov] = da.flip(image, axis=1)
            c_coords_dict[fov] = c_coords

        height, width = self.fov_locs["y1"].max(), self.fov_locs["x1"].max()

        if labels:
            stitched_image = da.zeros(shape=(height, width), dtype=image.dtype)
            stitched_image = xr.DataArray(stitched_image, dims=("y", "x"))
        else:
            c_coords = list(set.union(*[set(names) for names in c_coords_dict.values()]))

            stitched_image = da.zeros(shape=(len(c_coords), height, width), dtype=image.dtype)
            stitched_image = xr.DataArray(stitched_image, dims=("c", "y", "x"), coords={"c": c_coords})

        for fov, im in fov_images.items():
            xmin, xmax = self.fov_locs.loc[fov, "x0"], self.fov_locs.loc[fov, "x1"]
            ymin, ymax = self.fov_locs.loc[fov, "y0"], self.fov_locs.loc[fov, "y1"]

            if self.flip_image:
                y_slice, x_slice = slice(height - ymax, height - ymin), slice(width - xmax, width - xmin)
            else:
                y_slice, x_slice = slice(ymin, ymax), slice(xmin, xmax)

            if labels:
                stitched_image[y_slice, x_slice] = im[0]
            else:
                stitched_image.loc[{"c": c_coords_dict[fov], "y": y_slice, "x": x_slice}] = im

                if len(c_coords_dict[fov]) < len(c_coords):
                    log.warning(f"Missing channels ({len(c_coords) - len(c_coords_dict[fov])}) for FOV {fov}")

        if labels:
            return Labels2DModel.parse(stitched_image, **image_models_kwargs)
        else:
            return Image2DModel.parse(stitched_image, c_coords=c_coords, **image_models_kwargs)

    def _set_fov_locs_bounding_boxes(self, shape: tuple[int, int]) -> None:
        self.fov_locs["xmin"] = self.fov_locs["x"]
        self.fov_locs["xmax"] = self.fov_locs["x"] + shape[1]

        self.fov_locs["ymin"] = self.fov_locs["y"] - shape[0] * self.fov_shift
        self.fov_locs["ymax"] = self.fov_locs["y"] + shape[0] * (1 - self.fov_shift)

        for dim in ["x", "y"]:
            origin = self.fov_locs[f"{dim}min"].min()
            self.fov_locs[f"{dim}0"] = (self.fov_locs[f"{dim}min"] - origin).round().astype(int)
            self.fov_locs[f"{dim}1"] = (self.fov_locs[f"{dim}max"] - origin).round().astype(int)

    def _read_fov_locs(self) -> pd.DataFrame:
        fov_file = self.path / f"{self.dataset_id}_fov_positions_file.csv"

        if not fov_file.exists():
            fov_file = self.path / f"{self.dataset_id}_fov_positions_file.csv.gz"

        assert fov_file.exists(), f"Missing field of view file: {fov_file}"

        fov_locs = pd.read_csv(fov_file)
        fov_locs.columns = fov_locs.columns.str.lower()

        valid_keys = [
            ["fov", "x_global_px", "y_global_px"],
            ["fov", "x_mm", "y_mm"],
            ["fov", "x_global_mm", "y_global_mm"],
        ]
        mm_to_pixels = 1e3 / COSMX_PIXEL_SIZE  # conversion factor from mm to pixels for CosMx

        for (fov_key, x_key, y_key), scale_factor in zip(valid_keys, [1, mm_to_pixels, mm_to_pixels]):
            if not np.isin([fov_key, x_key, y_key], fov_locs.columns).all():  # try different column names
                continue

            fov_locs.index = fov_locs[fov_key]
            fov_locs["x"] = fov_locs[x_key] * scale_factor
            fov_locs["y"] = fov_locs[y_key] * scale_factor

            return fov_locs

        raise ValueError(
            f"The FOV positions file must contain one of the following sets of columns: {', or '.join(list(map(str, valid_keys)))}"
        )

    def _get_global_cell_id(self, df: pd.DataFrame) -> pd.Series:
        max_cell_id = df["cell_ID"].max()

        if hasattr(self, "max_cell_id"):
            assert max_cell_id == self.max_cell_id, (
                f"Expected max cell ID to be {self.max_cell_id}, but got {max_cell_id}."
            )
        self.max_cell_id = max_cell_id

        return df["fov"] * (self.max_cell_id + 1) * (df["cell_ID"] > 0) + df["cell_ID"]

    def _read_csv_gz(self, name: str) -> pd.DataFrame:
        for extension in [".gz", ""]:
            file = self.path / f"{name}{extension}"
            if file.exists():
                return pd.read_csv(file)
        raise FileNotFoundError(f"Input file not found: {self.path / f'{name}.gz'}")


def _infer_dataset_id(path: Path, dataset_id: str | None) -> str:
    if isinstance(dataset_id, str):
        return dataset_id

    for suffix in [".csv", ".csv.gz"]:
        counts_files = list(path.rglob(f"[!\\.]*_fov_positions_file{suffix}"))

        if len(counts_files) == 1:
            found = re.match(rf"(.*)_fov_positions_file{suffix}", counts_files[0].name)
            if found:
                return found.group(1)

    raise ValueError("Could not infer `dataset_id` from the name of the transcript file. Please specify it manually.")


def _read_fov_image(
    morphology_path: Path, protein_path: Path | None, morphology_coords: list[str], **imread_kwargs
) -> tuple[da.Array, list[str]]:
    image = imread(morphology_path, **imread_kwargs)

    protein_names = []
    if protein_path is not None:
        protein_image, protein_names = _read_protein_fov(protein_path)
        image = da.concatenate([image, protein_image], axis=0)

    return image, _deduplicate_names(morphology_coords + protein_names)


def _read_protein_fov(protein_dir: Path) -> tuple[da.Array, list[str]]:
    images_paths = list(protein_dir.rglob("*.[Tt][Ii][Ff]"))
    protein_image = da.concatenate([imread(image_path) for image_path in images_paths], axis=0)
    channel_names = [_get_protein_name(image_path) for image_path in images_paths]

    return protein_image, channel_names


def _find_matching_fov_file(images_dir: Path, fov: str | int) -> Path:
    assert isinstance(fov, int), "Expected `fov` to be an integer"

    pattern = re.compile(rf".*_F0*{fov}\.[Tt][Ii][Ff]")
    fov_files = [file for file in images_dir.rglob("*") if pattern.match(file.name)]

    assert len(fov_files), f"No file matches the pattern {pattern} inside {images_dir}"
    assert len(fov_files) == 1, f"Multiple files match the pattern {pattern}: {', '.join(map(str, fov_files))}"

    return fov_files[0]


def _find_dir(path: Path, name: str):
    if (path / name).is_dir():
        return path / name

    paths = list(path.rglob(f"**/{name}"))
    assert len(paths) == 1, f"Found {len(paths)} path(s) with name {name} inside {path}"

    return paths[0]


def _get_morphology_coords(images_dir: Path) -> list[str]:
    images_paths = list(images_dir.glob("*.[Tt][Ii][Ff]"))
    assert len(images_paths) > 0, f"Expected to find images inside {images_dir}"

    with tifffile.TiffFile(images_paths[0]) as tif:
        description = tif.pages[0].description

        substrings = re.findall(r'"BiologicalTarget": "(.*?)",', description)
        channels = re.findall(r'"ChannelId": "(.*?)",', description)
        channel_order = list(re.findall(r'"ChannelOrder": "(.*?)",', description)[0])

        channels: list[str] = [substrings[channels.index(x)] if x in channels else x for x in channel_order]
        return [channel.replace("/", ".") for channel in channels]


def _get_protein_name(image_path: Path) -> str:
    with tifffile.TiffFile(image_path) as tif:
        description = tif.pages[0].description
        substrings = re.findall(r'"DisplayName": "(.*?)",', description)
        return substrings[0].replace("/", ".")



from torch import nn


class HOPTIMUSFeatures(nn.Module):
    def __init__(self):
        super().__init__()

        try:
            import timm
        except ImportError:
            raise ImportError(
                "Using the 'hoptimus0' model for inference requires installing the timm dependency, e.g. via `pip install timm`"
            )

        self.model = timm.create_model("hf_hub:bioptimus/H-optimus-0", pretrained=True)
        self.model.eval()
        from torchvision import transforms

        self.transform = transforms.Compose([
            transforms.Normalize(mean=(0.707223, 0.578729, 0.703617), std=(0.211883, 0.230117, 0.177517)),
        ])

    def forward(self, x):
        return self.model(self.transform(x))



import logging
from math import ceil

import dask.dataframe as dd
import geopandas as gpd
import numpy as np
from shapely.geometry import GeometryCollection, MultiPolygon, Polygon, box
from spatialdata import SpatialData
from spatialdata.models import ShapesModel, SpatialElement
from xarray import DataArray, DataTree

from ..constants import SopaKeys
from ..shapes import to_valid_polygons
from ..utils import add_spatial_element, copy_transformations, to_intrinsic

log = logging.getLogger(__name__)


class Patches1D:
    def __init__(
        self,
        xmin: float | int,
        xmax: float | int,
        patch_width: float | int,
        patch_overlap: float | int,
        tight: bool,
        int_coords: bool,
    ):
        self.xmin, self.xmax = xmin, xmax
        self.delta = self.xmax - self.xmin

        self.patch_width = int(xmax - xmin + 1) if patch_width == float("inf") else patch_width
        self.patch_overlap = patch_overlap
        self.int_coords = int_coords

        self._count = self.count()
        if tight:
            self.patch_width = self.tight_width()
            assert self._count == self.count(), (
                f"Invalid patching with {self.delta=}, {self.patch_width=} and {self.patch_overlap=}"
            )

    def count(self):
        if self.patch_width >= self.delta:
            return 1
        return ceil((self.delta - self.patch_overlap) / (self.patch_width - self.patch_overlap))

    def update(self, patch_width):
        self.patch_width = patch_width
        assert self._count == self.count()

    def tight_width(self):
        width = (self.delta + (self._count - 1) * self.patch_overlap) / self._count
        return ceil(width) if self.int_coords else width + 1e-4  # add an epsilon to avoid floating point errors (#214)

    def __getitem__(self, i):
        start_delta = i * (self.patch_width - self.patch_overlap)
        x0, x1 = self.xmin + start_delta, self.xmin + start_delta + self.patch_width

        return [int(x0), int(x1)] if self.int_coords else [x0, x1]


class Patches2D:
    """
    Compute 2D-patches with overlaps. This can be done on an image or a DataFrame.

    Attributes:
        geo_df (GeoDataFrame): GeoPandas dataframe representing the patches
        bboxes (np.ndarray): Array of shape `(n_patches, 4)` containing the (xmin, ymin, xmax, ymax) coordinates of the patches bounding boxes
    """

    bboxes: np.ndarray
    geo_df: gpd.GeoDataFrame
    roi: Polygon | MultiPolygon | gpd.GeoDataFrame | None

    def __init__(
        self,
        sdata: SpatialData,
        element: SpatialElement | str,
        patch_width: float | int | None,
        patch_overlap: float | int = 50,
        roi_key: str | None = SopaKeys.ROI,
        use_roi_centroids: bool = False,
    ):
        """
        Args:
            sdata: A `SpatialData` object
            element: SpatialElement or name of the element on with patches will be made (image or points)
            patch_width: Width of the patches (in the unit of the coordinate system of the element). If `None`, only one patch will be created
            patch_overlap: Overlap width between the patches
            roi_key: Optional name of the shapes that need to touch the patches. Patches that do not touch any shape will be ignored. If `None`, all patches will be used.
            use_roi_centroids: If `True`, the ROI will be computed from the centroids of the shapes in `roi_key`. If `False`, the ROI will be computed from the shapes themselves.
        """
        patch_width = float("inf") if (patch_width is None or patch_width == -1) else patch_width

        assert patch_width > patch_overlap, f"Argument {patch_width=} must be greater than {patch_overlap=}"

        self.sdata = sdata
        self.element = sdata[element] if isinstance(element, str) else element
        self.original_element = self.element  # in case the element is a DataTree

        if isinstance(self.element, DataTree):
            self.element = next(iter(self.element["scale0"].values()))

        if isinstance(self.element, DataArray):
            xmin, ymin = 0, 0
            xmax, ymax = len(self.element.coords["x"]), len(self.element.coords["y"])
            tight, int_coords = False, True
        elif isinstance(self.element, dd.DataFrame):
            xmin, ymin = self.element.x.min().compute(), self.element.y.min().compute()
            xmax, ymax = self.element.x.max().compute(), self.element.y.max().compute()
            tight, int_coords = True, False
        else:
            raise TypeError(f"Invalid element type: {type(self.element)}")

        self.patch_x = Patches1D(xmin, xmax, patch_width, patch_overlap, tight, int_coords)
        self.patch_y = Patches1D(ymin, ymax, patch_width, patch_overlap, tight, int_coords)

        self.roi = None
        assert roi_key is None or roi_key in sdata.shapes or roi_key == SopaKeys.ROI, f"Invalid {roi_key=}"

        if roi_key is not None and roi_key in sdata.shapes:
            roi_geo_df = to_intrinsic(sdata, sdata[roi_key], self.original_element)
            self.roi = _get_roi(roi_geo_df, use_roi_centroids)

        self.geo_df = self._init_patches(use_roi_centroids)
        self.bboxes = np.array(self.geo_df[SopaKeys.BOUNDS].to_list())

    def _init_patches(self, use_roi_centroids: bool) -> gpd.GeoDataFrame:
        data = {
            "geometry": [],
            SopaKeys.PATCHES_ILOCS: [],
            SopaKeys.BOUNDS: [],
        }

        for i in range(self.patch_x._count * self.patch_y._count):
            iy, ix = divmod(i, self.patch_x._count)
            bounds = self._bbox_iloc(ix, iy)
            patch = box(*bounds)

            data["geometry"].append(patch)
            data[SopaKeys.PATCHES_ILOCS].append((ix, iy))
            data[SopaKeys.BOUNDS].append(bounds)

        geo_df = gpd.GeoDataFrame(data)

        if self.roi is not None:
            if use_roi_centroids:
                indices = gpd.sjoin(geo_df, self.roi).index.unique()
                geo_df = geo_df.loc[indices]
            else:
                geo_df.geometry = geo_df.geometry.intersection(self.roi)
                geo_df = geo_df[~geo_df.is_empty]

        geo_df = to_valid_polygons(geo_df, simple_polygon=False)
        geo_df = geo_df.reset_index(drop=True)

        assert len(geo_df), "No valid patches found inside the provided region of interest."

        return ShapesModel.parse(geo_df, transformations=copy_transformations(self.element))

    def __repr__(self):
        return f"{self.__class__.__name__} object with {len(self)} patches"

    @property
    def shape(self) -> tuple[int, int]:
        return (self.patch_y._count, self.patch_x._count)

    @property
    def is_full_slide(self) -> bool:
        return self.roi is None and len(self) == 1

    def centroids(self) -> np.ndarray:
        """Get the centroids of the patches as a numpy array of shape `(n_patches, 2)`.
        This doesn't take into account the ROI, so it may return centroids outside the ROI."""
        x = (self.bboxes[:, 0] + self.bboxes[:, 2]) / 2
        y = (self.bboxes[:, 1] + self.bboxes[:, 3]) / 2

        return np.column_stack((x, y))

    def _bbox_iloc(self, ix: int, iy: int) -> list[int]:
        """Coordinates of the rectangle bounding box of the patch at the given indices

        Args:
            ix: Patch index in the x-axis
            iy: Patch indes in the y-axis

        Returns:
            A list `[xmin, ymin, xmax, ymax]` representing the bounding box of the patch
        """
        xmin, xmax = self.patch_x[ix]
        ymin, ymax = self.patch_y[iy]
        return [xmin, ymin, xmax, ymax]

    def __len__(self):
        """Number of patches"""
        return len(self.geo_df)

    def add_shapes(self, key_added: str | None = None, overwrite: bool = True) -> gpd.GeoDataFrame:
        key_added = key_added or SopaKeys.PATCHES

        add_spatial_element(self.sdata, key_added, self.geo_df, overwrite=overwrite)

        log.info(f"Added {len(self.geo_df)} patch(es) to sdata['{key_added}']")

        return self.geo_df


def _get_roi(geo_df: gpd.GeoDataFrame, use_roi_centroids: bool) -> Polygon | MultiPolygon | gpd.GeoDataFrame:
    """Merge all geometries into a single region-of-interest"""
    if use_roi_centroids:
        return geo_df.centroid.to_frame()

    roi = geo_df.geometry.union_all()

    if isinstance(roi, GeometryCollection):
        _previous_area = roi.area

        roi = MultiPolygon([geom for geom in roi.geoms if isinstance(geom, Polygon)])
        if roi.area < _previous_area * 0.999:
            raise ValueError("ROI is a GeometryCollection that could not be simplified into polygons.")

    assert isinstance(roi, (Polygon, MultiPolygon)), f"Invalid ROI type: {type(roi)}. Must be Polygon or MultiPolygon"

    return roi



from .utils import (
    get_cache_dir,
    delete_cache,
    get_boundaries,
    get_spatial_element,
    to_intrinsic,
    get_spatial_image,
    get_intensities,
    add_spatial_element,
    get_transcripts_patches_dirs,
    get_feature_key,
    delete_transcripts_patches_dirs,
    set_sopa_attrs,
    set_boundaries_attrs,
    copy_transformations,
)
from .annotation import preprocess_fluo, higher_z_score, tangram_annotate
from .image import (
    get_channel_names,
    validated_channel_names,
    scale_dtype,
    ensure_string_channel_names,
    is_valid_c_coords,
    assert_is_integer_dtype,
    resize_numpy,
)



import logging
import warnings
from pathlib import Path

import dask.dataframe as dd
import geopandas as gpd
import numpy as np
import pandas as pd
import spatialdata
from spatialdata import SpatialData
from spatialdata.models import SpatialElement
from spatialdata.transformations import (
    Affine,
    BaseTransformation,
    Identity,
    Sequence,
    get_transformation,
    get_transformation_between_coordinate_systems,
)
from xarray import DataArray, DataTree

from .. import settings
from ..constants import ATTRS_KEY, SopaAttrs, SopaFiles, SopaKeys

log = logging.getLogger(__name__)


def set_boundaries_attrs(sdata: SpatialData, key: str):
    """Sets the boundaries attribute in the SpatialData object.

    Args:
        sdata: A SpatialData object.
        key: Key of the shapes containing the cell boundaries.
    """
    assert key in sdata.shapes, f"No '{key}' found in sdata.shapes"
    sdata.attrs[SopaAttrs.BOUNDARIES] = key
    _save_attrs(sdata)


def get_boundaries(
    sdata: SpatialData,
    return_key: bool = False,
    warn: bool = False,
    key: str | None = None,
    table_key: str | None = None,
) -> gpd.GeoDataFrame | tuple[str, gpd.GeoDataFrame] | None:
    """Gets cell segmentation boundaries of a SpatialData object after running Sopa.

    Args:
        sdata: A SpatialData object
        return_key: Whether to return the key of the shapes or not.
        warn: If `True`, prints a warning if no boundary is found. Else, raises an error.
        key: A valid `shapes_key` or None.
        table_key: Name of the table used to find the corresponding boundaries.

    Returns:
        A `GeoDataFrame` containing the boundaries, or a tuple `(shapes_key, geo_df)`
    """
    assert key is None or table_key is None, "Provide only one of `key` or `table_key`"

    if table_key is not None:
        key = sdata.tables[table_key].uns[ATTRS_KEY]["region"]
        assert isinstance(key, str)
        return get_spatial_element(sdata.shapes, key=key, return_key=return_key)

    key = key or sdata.attrs.get(SopaAttrs.BOUNDARIES)

    if key is not None:
        return get_spatial_element(sdata.shapes, key=key, return_key=return_key)

    VALID_BOUNDARIES = [
        SopaKeys.PROSEG_BOUNDARIES,
        SopaKeys.BAYSOR_BOUNDARIES,
        SopaKeys.STARDIST_BOUNDARIES,
        SopaKeys.COMSEG_BOUNDARIES,
        SopaKeys.CELLPOSE_BOUNDARIES,
    ]
    for key in VALID_BOUNDARIES:
        res = _try_get_boundaries(sdata, key, return_key)
        if res is not None:
            return res

    error_message = "sdata object has no valid segmentation boundary. Consider running Sopa segmentation first."

    if not warn:
        raise ValueError(error_message)

    log.warning(error_message)
    return (None, None) if return_key else None


def _try_get_boundaries(sdata: SpatialData, key: str, return_key: bool) -> gpd.GeoDataFrame | None:
    """Try to get a cell boundaries for a given `shapes_key`"""
    if key in sdata.shapes:
        return (key, sdata[key]) if return_key else sdata[key]


def to_intrinsic(
    sdata: SpatialData, element: SpatialElement | str, target_element: SpatialElement | str
) -> SpatialElement:
    """Transforms a `SpatialElement` into the intrinsic coordinate system of another `SpatialElement`

    Args:
        sdata: A SpatialData object
        element: `SpatialElement` to transform, or its key. We recommend it to choose a vector element (for instance, points or shapes).
        target_element: `SpatialElement` of the target coordinate system, or its key.

    Returns:
        The `element` with coordinates transformed to the intrinsic coordinate system of `target_element`.
    """
    element = sdata[element] if isinstance(element, str) else element
    target_element = sdata[target_element] if isinstance(target_element, str) else target_element

    for cs, transformation in get_transformation(element, get_all=True).items():
        if isinstance(transformation, Identity):
            target_transformations = get_transformation(target_element, get_all=True)
            if isinstance(target_transformations.get(cs), Identity):
                return element  # no transformation needed
            break

    try:
        transformation = get_transformation_between_coordinate_systems(sdata, element, target_element)
    except:
        transformations1 = get_transformation(element, get_all=True)
        transformations2 = get_transformation(target_element, get_all=True)

        common_keys = list(set(transformations1.keys()) & set(transformations2.keys()))

        if not common_keys:
            raise ValueError("No common coordinate system found between the two elements")

        cs = "global" if "global" in common_keys else common_keys.pop()

        transformation = Sequence([transformations1[cs], transformations2[cs].inverse()])

    return spatialdata.transform(element, transformation=transformation, maintain_positioning=True)


def _make_affine_2d(affine: Affine, remove_name: str = "z") -> Affine:
    X = affine.matrix
    i, j = affine.input_axes.index(remove_name), affine.output_axes.index(remove_name)

    assert X[j, i] == 1 and len(X[:, i].nonzero()[0]) == 1, "Cannot reduce 3D affine to 2D"

    return Affine(
        np.delete(np.delete(X, i, axis=1), j, axis=0),
        input_axes=[name for name in affine.input_axes if name != remove_name],
        output_axes=[name for name in affine.output_axes if name != remove_name],
    )


def _has_3d_name(affine: Affine, name: str) -> bool:
    in_input, in_output = name in affine.input_axes, name in affine.output_axes
    assert in_input == in_output, f"Affine transformation has {name=} in only one of input/output axes: {affine}"
    return in_input


def _ensure_2d_transformation(
    transformation: BaseTransformation | dict[str, BaseTransformation],
) -> BaseTransformation | dict[str, BaseTransformation]:
    if isinstance(transformation, dict):
        return {cs: _ensure_2d_transformation(t) for cs, t in transformation.items()}

    if isinstance(transformation, Affine):
        if _has_3d_name(transformation, "z"):
            return _make_affine_2d(transformation, remove_name="z")
        elif _has_3d_name(transformation, "c"):
            return _make_affine_2d(transformation, remove_name="c")
        return transformation

    elif isinstance(transformation, Sequence):
        return Sequence([_ensure_2d_transformation(t) for t in transformation.transformations])

    return transformation


def copy_transformations(element: SpatialElement, ensure_2d: bool = True) -> BaseTransformation:
    transformation = get_transformation(element, get_all=True).copy()
    return _ensure_2d_transformation(transformation) if ensure_2d else transformation


def get_feature_key(points: dd.DataFrame | SpatialData, raise_error: bool = False) -> str:
    """Get the feature key of a transcript dataframe"""
    if isinstance(points, SpatialData):
        if SopaKeys.TRANSCRIPTS_PATCHES in points.shapes:
            points_key = points[SopaKeys.TRANSCRIPTS_PATCHES][SopaKeys.POINTS_KEY].iloc[0]
        else:
            points_key = points.attrs.get(SopaAttrs.TRANSCRIPTS)
            assert points_key is not None, "No transcripts found in the SpatialData object"

        points = points[points_key]

    assert isinstance(points, dd.DataFrame), "points must be a Dask DataFrame"

    feature_key = points.attrs.get(ATTRS_KEY, {}).get("feature_key")

    if raise_error and feature_key is None:
        raise ValueError(
            f"No gene column name found in points['{ATTRS_KEY}']['feature_key']. Provide the `gene_column` argument."
        )

    return feature_key


def get_intensities(sdata: SpatialData, table_key: str = SopaKeys.TABLE) -> pd.DataFrame | None:
    """Gets the intensity dataframe of shape `n_obs x n_channels`

    Args:
        sdata: A `SpatialData` object.
        table_key: Key of `sdata` containing to table from which intensities will be extracted.

    Returns:
        A pandas DataFrame containing the intensities, or `None` if no intensities are found.
    """
    assert table_key in sdata.tables, f"No '{table_key}' found in sdata.tables"

    adata = sdata.tables[table_key]

    if not adata.uns[SopaKeys.UNS_KEY][SopaKeys.UNS_HAS_INTENSITIES]:
        return None

    if adata.uns[SopaKeys.UNS_KEY][SopaKeys.UNS_HAS_TRANSCRIPTS]:
        return adata.obsm[SopaKeys.INTENSITIES_OBSM]

    return adata.to_df()


def get_spatial_element(
    element_dict: dict[str, SpatialElement],
    key: str | None = None,
    return_key: bool = False,
    as_spatial_image: bool = False,
) -> SpatialElement | tuple[str, SpatialElement]:
    """Gets an element from a SpatialData object.

    Args:
        element_dict: Dictionnary whose values are spatial elements (e.g., `sdata.images`).
        key: Optional element key. If `None`, returns the only element (if only one).
        return_key: Whether to also return the key of the element.
        as_spatial_image: Whether to return the element as a `SpatialImage` (if it is a `DataTree`)

    Returns:
        If `return_key` is False, only the element is returned, else a tuple `(element_key, element)`
    """
    assert len(element_dict), "No spatial element was found in the dict."

    if key is not None:
        assert key in element_dict, f"Spatial element '{key}' not found."
        return _return_element(element_dict, key, return_key, as_spatial_image)

    assert len(element_dict) > 0, (
        "No spatial element found. Provide an element key to denote which element you want to use."
    )
    assert len(element_dict) == 1, (
        f"Multiple valid elements found: {', '.join(element_dict.keys())}. Provide an element key to denote which element you want to use."
    )

    key = next(iter(element_dict.keys()))

    return _return_element(element_dict, key, return_key, as_spatial_image)


def get_spatial_image(
    sdata: SpatialData,
    key: str | None = None,
    return_key: bool = False,
    valid_attr: str = SopaAttrs.CELL_SEGMENTATION,
) -> DataArray | tuple[str, DataArray]:
    """Gets a DataArray from a SpatialData object (if the image has multiple scale, the `scale0` is returned)

    Args:
        sdata: SpatialData object.
        key: Optional image key. If `None`, returns the only image (if only one), or tries to find an image with `valid_attr`.
        return_key: Whether to also return the key of the image.
        valid_attr: Attribute that the image must have to be considered valid.

    Returns:
        If `return_key` is False, only the image is returned, else a tuple `(image_key, image)`
    """
    return get_spatial_element(
        sdata.images,
        key=key or sdata.attrs.get(valid_attr),
        return_key=return_key,
        as_spatial_image=True,
    )


def _return_element(
    element_dict: dict[str, SpatialElement], key: str, return_key: bool, as_spatial_image: bool
) -> SpatialElement | tuple[str, SpatialElement]:
    element = element_dict[key]

    if as_spatial_image and isinstance(element, DataTree):
        element = next(iter(element["scale0"].values()))

    return (key, element) if return_key else element


def add_spatial_element(
    sdata: SpatialData,
    element_name: str,
    element: SpatialElement,
    overwrite: bool = True,
):
    assert isinstance(element_name, str)
    assert overwrite or element_name not in sdata._shared_keys, (
        f"Trying to add {element_name=} but it is already existing and {overwrite=}"
    )

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message=".*already exists. Overwriting it in-memory.")
        sdata[element_name] = element

    if sdata.is_backed() and settings.auto_save_on_disk:
        try:
            sdata.write_element(element_name, overwrite=overwrite)
        except Exception as e:
            if overwrite:  # force writing the element
                sdata.delete_element_from_disk(element_name)
                sdata.write_element(element_name, overwrite=overwrite)
            else:
                raise ValueError(f"Error while saving {element_name} on disk with {overwrite=}: {e}")


def set_sopa_attrs(
    sdata: SpatialData,
    cell_segmentation_key: str | None = None,
    tissue_segmentation_key: str | None = None,
    transcripts_key: str | None = None,
    boundaries_key: str | None = None,
    bins_table_key: str | None = None,
):
    """Stores in the `SpatialData` object the keys of the main elements used in Sopa.
    This allows Sopa to retreive with elements should be used for each operation.

    !!! info
        The attrs are already stored in `sdata.attrs` when reading data with `sopa.io`.
        Use this function only if you already stored on disk a SpatialData object without the attrs (with `sopa<2.0.0`).

    Args:
        sdata: A `SpatialData` object.
        cell_segmentation_key: Name of the image to be used for cell segmentation (highest resolution image).
        tissue_segmentation_key: Name of the image to be used for tissue segmentation (medium/low resolution image).
        transcripts_key: Name of the points containing the transcripts.
        boundaries_key: Name of the shapes containing the cell boundaries.
        bins_table_key: Name of the table containing the bins (e.g., for Visium HD data).
    """
    if cell_segmentation_key is not None:
        assert cell_segmentation_key in sdata.images
        sdata.attrs[SopaAttrs.CELL_SEGMENTATION] = cell_segmentation_key

    if tissue_segmentation_key is not None:
        assert tissue_segmentation_key in sdata.images
        sdata.attrs[SopaAttrs.TISSUE_SEGMENTATION] = tissue_segmentation_key

    if transcripts_key is not None:
        assert transcripts_key in sdata.points
        sdata.attrs[SopaAttrs.TRANSCRIPTS] = transcripts_key

    if boundaries_key is not None:
        assert boundaries_key in sdata.shapes
        sdata.attrs[SopaAttrs.BOUNDARIES] = boundaries_key

    if bins_table_key is not None:
        assert bins_table_key in sdata.tables
        sdata.attrs[SopaAttrs.BINS_TABLE] = bins_table_key

    _save_attrs(sdata)


def _save_attrs(sdata: SpatialData):
    if sdata.is_backed() and settings.auto_save_on_disk:
        try:
            sdata.write_attrs()
        except Exception as e:
            raise ValueError(f"Error while saving attrs on disk: {e}")


HOME_CACHE_DIR = Path.home() / SopaFiles.SOPA_CACHE_DIR


def get_cache_dir(sdata: SpatialData) -> Path:
    """Get the cache directory for a SpatialData object.

    Args:
        sdata: A `SpatialData` object.

    Returns:
        A `Path` to the cache directory.
    """
    if sdata.is_backed():  # inside the zarr directory
        cache_dir = sdata.path.resolve() / SopaFiles.SOPA_CACHE_DIR
    elif SopaAttrs.UID in sdata.attrs:  # existing cache in the home directory
        cache_dir = HOME_CACHE_DIR / sdata.attrs[SopaAttrs.UID]
    else:  # create a new cache directory in the home directory
        import uuid

        uid = str(uuid.uuid4())
        sdata.attrs[SopaAttrs.UID] = uid
        cache_dir = HOME_CACHE_DIR / str(uid)

    cache_dir.mkdir(exist_ok=True, parents=True)

    return cache_dir


def delete_cache(sdata: SpatialData | None = None) -> None:
    """Delete the cache directory (the entire cache, or the cache of one specific SpatialData object).

    Args:
        sdata: The SpatialData object whose cache is to be deleted. If None, the entire cache is deleted.
    """
    import shutil

    if sdata is not None:
        cache_dir = get_cache_dir(sdata)
        shutil.rmtree(cache_dir)
        return

    for sub_dir in list(HOME_CACHE_DIR.iterdir()):
        if sub_dir.is_dir():
            shutil.rmtree(sub_dir)


def get_transcripts_patches_dirs(sdata: SpatialData, index: int | None = None) -> list[Path] | Path:
    """Get the list of directories containing the transcript patches

    Args:
        sdata: A `SpatialData` object containing the transcript patches.

    Returns:
        If `index` is None, a list of `Path` to the directories containing the transcript patches. Else, the `Path` to the directory of the patch with the given index.
    """
    assert SopaKeys.TRANSCRIPTS_PATCHES in sdata.shapes, "Transcript patches not found in the SpatialData object"

    cache_dir = get_cache_dir(sdata) / SopaFiles.TRANSCRIPT_CACHE_DIR

    if index is not None:
        assert index in sdata.shapes[SopaKeys.TRANSCRIPTS_PATCHES].index, f"Patch index {index} not found"
        return cache_dir / str(index)

    return [cache_dir / str(index) for index in sdata.shapes[SopaKeys.TRANSCRIPTS_PATCHES].index]


def delete_transcripts_patches_dirs(sdata: SpatialData):
    """Delete the cache directories containing the transcript patches (for instance, for Baysor or ComSeg)

    Args:
        sdata: A `SpatialData` object.
    """
    import shutil

    for patch_dir in get_transcripts_patches_dirs(sdata):
        shutil.rmtree(patch_dir)



import logging
from collections.abc import Iterable
from functools import partial
from pathlib import Path
from typing import Callable

import geopandas as gpd
import numpy as np
import pandas as pd
import spatialdata
from scipy.ndimage import gaussian_filter
from shapely.geometry import Polygon, box
from skimage import exposure
from spatialdata import SpatialData
from spatialdata.models import ShapesModel

from .. import settings, shapes
from ..constants import SopaKeys
from ..utils import add_spatial_element, copy_transformations, get_spatial_image

log = logging.getLogger(__name__)


class StainingSegmentation:
    def __init__(
        self,
        sdata: SpatialData,
        method: Callable,
        channels: list[str] | str | None,
        image_key: str | None = None,
        min_area: int | float = 0,
        clip_limit: float = 0.2,
        clahe_kernel_size: int | Iterable[int] | None = None,
        gaussian_sigma: float = 1,
        min_patch_size: int = 10,
    ):
        """Generalized staining-based segmentation class

        Args:
            sdata: A `SpatialData` object
            method: A segmentation `callable` whose input is an image of shape `(C, Y, X)` and output is a cell mask of shape `(Y, X)`. Each mask value `>0` represent a unique cell ID. Such callables can be found in `sopa.segmentation.methods`.
            channels: One or a list of channel names used for segmentation. If only one channel is provided, the image given to the `method` will be of shape `(1, Y, X)`. None assumes RGB image.
            image_key: Optional key of `sdata` containing the image (no needed if there is only one image)
            min_area: Minimum area (in pixels^2) for a cell to be kept
            clip_limit: Parameter for skimage.exposure.equalize_adapthist (applied before running cellpose)
            clahe_kernel_size: Parameter for skimage.exposure.equalize_adapthist (applied before running cellpose)
            gaussian_sigma: Parameter for scipy gaussian_filter (applied before running cellpose)
            min_patch_size: Minimum patch size (in pixels) for both width and height. Patches smaller than this will be skipped to avoid segmentation errors.
        """
        assert SopaKeys.PATCHES in sdata.shapes, "Run `sopa.make_image_patches` before running segmentation"

        self.patches_gdf: gpd.GeoDataFrame = sdata[SopaKeys.PATCHES]
        self.method = method

        self.min_area = min_area
        self.clip_limit = clip_limit
        self.clahe_kernel_size = clahe_kernel_size
        self.gaussian_sigma = gaussian_sigma
        self.min_patch_size = min_patch_size

        self.image_key, self.image = get_spatial_image(sdata, key=image_key, return_key=True)

        image_channels = self.image.coords["c"].values

        if channels is None:
            assert len(image_channels) == 3, "No channels were provided. This is only possible for RGB images."
            self.channels = image_channels
        else:
            self.channels = [channels] if isinstance(channels, str) else channels
            assert np.isin(channels, image_channels).all(), (
                f"Channel names must be a subset of: {', '.join(image_channels)}"
            )

    def _run_patch(self, patch: Polygon) -> gpd.GeoDataFrame:
        """Run segmentation on one patch

        Args:
            patch: Patch, represented as a `shapely` polygon

        Returns:
            A list of cells, represented as polygons
        """
        bounds = [int(x) for x in patch.bounds]

        patch_width, patch_height = bounds[2] - bounds[0], bounds[3] - bounds[1]
        if min(patch_width, patch_height) < self.min_patch_size:
            log.info(f"Skipping patch with too small dimensions ({patch_width}, {patch_height})")
            return gpd.GeoDataFrame(geometry=[])

        image = self.image.sel(
            c=self.channels,
            x=slice(bounds[0], bounds[2]),
            y=slice(bounds[1], bounds[3]),
        ).values

        assert np.issubdtype(image.dtype, np.integer), (
            f"Invalid image type {image.dtype}. Transform it to an integer dtype, e.g. `np.uint8`."
        )

        if self.gaussian_sigma > 0:
            image = np.stack([gaussian_filter(c, sigma=self.gaussian_sigma) for c in image])
        if self.clip_limit > 0:
            image = np.stack([
                exposure.equalize_adapthist(
                    c,
                    clip_limit=self.clip_limit,
                    kernel_size=self.clahe_kernel_size,
                )
                for c in image
            ])

        if patch.area < box(*bounds).area:
            mask = shapes.rasterize(patch, image.shape[1:], bounds)
            image = _channels_average_within_mask(image, mask)

        cells = shapes.vectorize(self.method(image))
        cells.geometry = cells.translate(*bounds[:2])

        return cells[cells.area >= self.min_area] if self.min_area > 0 else cells

    def write_patch_cells(self, patch_dir: str, patch_index: int, recover: bool = False):
        """Run segmentation on one patch, and save the result in a dedicated directory

        Args:
            patch_dir: Directory inside which segmentation results will be saved
            patch_index: Index of the patch on which to run segmentation. NB: the number of patches is `len(sdata['image_patches'])`
            recover: If `True`, the function will not run segmentation if the output file already exists
        """
        patch_dir: Path = Path(patch_dir)
        patch_dir.mkdir(parents=True, exist_ok=True)
        output_path = patch_dir / f"{patch_index}.parquet"

        if recover and output_path.exists():
            return

        patch = self.patches_gdf.geometry[patch_index]

        cells = self._run_patch(patch)
        cells.to_parquet(output_path)

    def write_patches_cells(self, patch_dir: str, recover: bool = False):
        """Run segmentation on all patches, and save the result in a dedicated directory

        Args:
            patch_dir: Directory inside which segmentation results will be saved
            recover: If `True`, the function will not run segmentation on already-segmented patches
        """
        functions = [
            partial(self.write_patch_cells, patch_dir, patch_index, recover)
            for patch_index in range(len(self.patches_gdf))
        ]

        if settings.parallelization_backend is not None and not len(spatialdata.get_dask_backing_files(self.image)):
            log.warning(
                "Having backed data is recommended when running staining-based segmentation with a parallelization backend. "
                "Consider saving your data on disk with `sdata.write(...)`, and opening it with `sdata = spatialdata.read_zarr(...)`."
            )

        settings._run_with_backend(functions)

    @classmethod
    def read_patches_cells(cls, patch_dir: str | list[str]) -> gpd.GeoDataFrame:
        """Read all patch segmentation results after running `write_patch_cells` on all patches

        Args:
            patch_dir: Directory provided when running `write_patch_cells` containing the `.parquet` files. For multi-step segmentation, provide a list of directories (one per step).

        Returns:
            A list of cells represented as `shapely` polygons
        """
        patch_dirs = patch_dir if isinstance(patch_dir, list) else [patch_dir]
        files = [f for patch_dir in patch_dirs for f in Path(patch_dir).iterdir() if f.suffix == ".parquet"]

        cells = pd.concat([gpd.read_parquet(file) for file in files]).reset_index(drop=True)

        log.info(f"Found {len(cells)} total cells")
        return cells

    @classmethod
    def add_shapes(cls, sdata: SpatialData, cells: gpd.GeoDataFrame, image_key: str, key_added: str):
        """Adding `shapely` polygon to the `SpatialData` object

        Args:
            sdata: A `SpatialData` object
            cells: `GeoDataFrame` containing the cell boundaries after segmentation
            image_key: Key of the image on which segmentation has been run
            shapes_key: Name to provide to the geodataframe to be created
        """
        image = get_spatial_image(sdata, image_key)

        cells.index = image_key + cells.index.astype(str)
        cells = ShapesModel.parse(cells, transformations=copy_transformations(image))

        add_spatial_element(sdata, key_added, cells)

        log.info(f"Added {len(cells)} cell boundaries in sdata['{key_added}']")


def _channels_average_within_mask(image: np.ndarray, mask: np.ndarray) -> np.ndarray:
    channels_average = (image * mask).sum(axis=(1, 2)) / mask.sum().clip(1)

    return image * mask + (1 - mask) * channels_average[:, None, None]



from .build import spatial_neighbors
from .distance import cells_to_groups, mean_distance, prepare_network
from .join import assign_transcript_to_cell, sjoin
from .morpho import geometrize_niches, niches_geometry_stats, vectorize_niches



"""
Copied from squidpy (to not have squidpy as a dependency)
Functions for building graphs from spatial coordinates.
"""

import logging
import warnings
from collections.abc import Iterable
from functools import partial
from itertools import chain

import numpy as np
from anndata import AnnData
from anndata.utils import make_index_unique
from scipy.sparse import SparseEfficiencyWarning, block_diag, csr_matrix, spmatrix
from scipy.spatial import Delaunay
from sklearn.metrics.pairwise import euclidean_distances
from spatialdata import SpatialData

from ..constants import SopaKeys

log = logging.getLogger(__name__)
__all__ = ["spatial_neighbors"]


def spatial_neighbors(
    adata: AnnData | SpatialData,
    radius: tuple[float, float] | None,
    library_key: str | None = None,
    percentile: float | None = None,
    set_diag: bool = False,
):
    """Create a Delaunay graph from spatial coordinates. This function comes from [squidpy](https://squidpy.readthedocs.io/en/latest/api/squidpy.gr.spatial_neighbors.html#squidpy.gr.spatial_neighbors).

    Args:
        adata: AnnData object
        radius: tuple that prunes the final graph to only contain edges in interval `[min(radius), max(radius)]`. If `None`, all edges are kept.
        library_key: Optional batch key in adata.obs
        percentile: Percentile of the distances to use as threshold.
        set_diag: Whether to set the diagonal of the spatial connectivities to `1.0`.
    """
    if isinstance(adata, SpatialData):
        adata = adata.tables[SopaKeys.TABLE]

    assert radius is None or len(radius) == 2, "Radius is expected to be a tuple (min_radius, max_radius)"

    log.info("Computing delaunay graph")

    if library_key is not None:
        assert adata.obs[library_key].dtype == "category"
        libs = adata.obs[library_key].cat.categories
        make_index_unique(adata.obs_names)
    else:
        libs = [None]

    _build_fun = partial(
        _spatial_neighbor,
        set_diag=set_diag,
        radius=radius,
        percentile=percentile,
    )

    if library_key is not None:
        mats: list[tuple[spmatrix, spmatrix]] = []
        ixs = []  # type: ignore[var-annotated]
        for lib in libs:
            ixs.extend(np.where(adata.obs[library_key] == lib)[0])
            mats.append(_build_fun(adata[adata.obs[library_key] == lib]))
        ixs = np.argsort(ixs)  # type: ignore[assignment] # invert
        Adj = block_diag([m[0] for m in mats], format="csr")[ixs, :][:, ixs]
        Dst = block_diag([m[1] for m in mats], format="csr")[ixs, :][:, ixs]
    else:
        Adj, Dst = _build_fun(adata)

    adata.obsp["spatial_connectivities"] = Adj
    adata.obsp["spatial_distances"] = Dst


def _spatial_neighbor(
    adata: AnnData,
    radius: float | tuple[float, float] | None = None,
    set_diag: bool = False,
    percentile: float | None = None,
) -> tuple[csr_matrix, csr_matrix]:
    coords = adata.obsm["spatial"]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", SparseEfficiencyWarning)
        Adj, Dst = _build_connectivity(
            coords,
            set_diag=set_diag,
        )

    if isinstance(radius, Iterable):
        minn, maxx = sorted(radius)[:2]  # type: ignore[var-annotated]
        mask = (Dst.data < minn) | (Dst.data > maxx)
        a_diag = Adj.diagonal()

        Dst.data[mask] = 0.0
        Adj.data[mask] = 0.0
        Adj.setdiag(a_diag)

    if percentile is not None:
        threshold = np.percentile(Dst.data, percentile)
        Adj[Dst > threshold] = 0.0
        Dst[Dst > threshold] = 0.0

    Adj.eliminate_zeros()
    Dst.eliminate_zeros()

    return Adj, Dst


def _build_connectivity(
    coords: np.ndarray,
    set_diag: bool = False,
) -> csr_matrix | tuple[csr_matrix, csr_matrix]:
    N = coords.shape[0]

    tri = Delaunay(coords)
    indptr, indices = tri.vertex_neighbor_vertices
    Adj = csr_matrix((np.ones_like(indices, dtype=np.float64), indices, indptr), shape=(N, N))

    # fmt: off
    dists = np.array(list(chain(*(
        euclidean_distances(coords[indices[indptr[i] : indptr[i + 1]], :], coords[np.newaxis, i, :])
        for i in range(N)
        if len(indices[indptr[i] : indptr[i + 1]])
    )))).squeeze()
    Dst = csr_matrix((dists, indices, indptr), shape=(N, N))
    # fmt: on

    # radius-based filtering needs same indices/indptr: do not remove 0s
    Adj.setdiag(1.0 if set_diag else Adj.diagonal())
    Dst.setdiag(0.0)

    return Adj, Dst


def _check_has_delaunay(adata: AnnData):
    message = " key not in adata.obsp, consider running the delaunay graph (e.g., `from sopa.spatial import spatial_neighbors; spatial_neighbors(adata, [0, 40])`)"
    assert "spatial_connectivities" in adata.obsp, "spatial_connectivities" + message
    assert "spatial_distances" in adata.obsp, "spatial_distances" + message



import logging
import math

import numpy as np
import pandas as pd
import scanpy as sc
from anndata import AnnData
from spatialdata import SpatialData

from ..constants import SopaKeys

log = logging.getLogger(__name__)


def preprocess_fluo(adata: AnnData) -> pd.DataFrame:
    """Preprocess fluorescence data. For each column $X$, we compute $asinh(\\frac{X}{5Q(0.2, X)})$ and apply standardization

    Args:
        adata: An `AnnData` object

    Returns:
        A dataframe of preprocessed channels intensities
    """
    df = adata.obsm[SopaKeys.INTENSITIES_OBSM] if SopaKeys.INTENSITIES_OBSM in adata.obsm else adata.to_df()

    divider = 5 * np.quantile(df, 0.2, axis=0)
    divider[divider == 0] = df.max(axis=0)[divider == 0]

    scaled = np.arcsinh(df / divider)
    return (scaled - scaled.mean(0)) / scaled.std(0)


def higher_z_score(adata: AnnData, marker_cell_dict: dict, cell_type_key: str = "cell_type"):
    """Simple channel-based segmentation using a marker-to-population dictionary

    Args:
        adata: An `AnnData` object
        marker_cell_dict: Dictionary whose keys are channels, and values are the corresponding populations.
        cell_type_key: Key of `adata.obs` where annotations will be stored
    """
    adata.obsm[SopaKeys.Z_SCORES] = preprocess_fluo(adata)

    markers, cell_types = list(marker_cell_dict.keys()), np.array(list(marker_cell_dict.values()))
    ct_indices = adata.obsm[SopaKeys.Z_SCORES][markers].values.argmax(1)

    adata.obs[cell_type_key] = cell_types[ct_indices]
    adata.uns[SopaKeys.UNS_KEY][SopaKeys.UNS_CELL_TYPES] = [cell_type_key]

    log.info(f"Annotation counts: {adata.obs[cell_type_key].value_counts()}")


def tangram_annotate(
    sdata: SpatialData | AnnData,
    adata_sc: AnnData,
    cell_type_key: str,
    reference_preprocessing: str | None = None,
    bag_size: int = 10_000,
    max_obs_reference: int = 10_000,
    density_prior: str = "uniform",
    clip_percentile: float = 0.95,
):
    """Tangram multi-level annotation. Tangram is run on multiple bags of cells to decrease the RAM usage.

    !!! info
        You need to install `tangram-sc` to use this function. You can install it via `pip install tangram-sc`.

    !!! info "Multi-level annotation"
        If multi-level annotation is used (see `cell_type_key` argument), Tangram is first run on level 0. Then, for each cell-type, Tangram is run again within its annotated subset to assign subtypes. This process continues recursively, with each level processed within the cells labeled at the previous level.

    Args:
        sdata: A `SpatialData` object containing an AnnData table, or the AnnData table itself.
        adata_sc: A scRNAseq annotated reference containing cell types in `adata_sc.obs[cell_type_key]`. If it has been pre-processed, you can provide the `reference_preprocessing` argument to apply the same preprocessing on the spatial data. If containing raw counts, the spatial table should also contain raw counts.
        cell_type_key: Key of `adata_sc.obs` containing the cell types. For multi-level annotation, provide other levels like such: if `cell_type_key = "ct"`, then `"ct_level1"` and `"ct_level2"` are the two next levels.
        reference_preprocessing: Preprocessing method that was already applied on the reference. Can be `"log1p"` (normalize_total + log1p) or `"normalized"` (just normalize_total). If provided, it will apply the same preprocessing on the spatial data before running Tangram. By default, no preprocessing is applied.
        bag_size: Size of each bag on which tangram will be run. Use smaller bags to lower the RAM usage. If `bag_size` is too small, it might create less consistent results across bags.
        max_obs_reference: Maximum number of cells used in `adata_sc` at each level. Decrease it to lower the RAM usage.
        density_prior: Density prior used in Tangram. Can be `"uniform"` or `"rna_count_based"`.
        clip_percentile: Percentile used to clip the probabilities before taking the maximum (to obtain hard cell-type labels from probabilities).
    """
    if isinstance(sdata, SpatialData):
        assert SopaKeys.TABLE in sdata.tables, f"No '{SopaKeys.TABLE}' table found in sdata.tables"
        ad_sp = sdata.tables[SopaKeys.TABLE]
    else:
        assert isinstance(sdata, AnnData), "sdata must be a SpatialData object or an AnnData object"
        ad_sp = sdata

    if reference_preprocessing is None and sum(adata.X.max() > 10 for adata in [ad_sp, adata_sc]) == 1:
        log.warning(
            "It seems that the spatial data is not pre-processed the same way as the reference, but `reference_preprocessing` is not set. "
            "Either provide the `reference_preprocessing` argument or pre-process both AnnData the same way before running Tangram."
        )

    MultiLevelAnnotation(
        ad_sp,
        adata_sc,
        cell_type_key,
        reference_preprocessing,
        bag_size,
        max_obs_reference,
        density_prior,
        clip_percentile,
    ).run()


class MultiLevelAnnotation:
    def __init__(
        self,
        ad_sp: AnnData,
        ad_sc: AnnData,
        cell_type_key: str,
        reference_preprocessing: str | None,
        bag_size: int,
        max_obs_reference: int,
        density_prior: str,
        clip_percentile: float,
    ):
        self.ad_sp = ad_sp
        self.ad_sc = ad_sc

        if not self.ad_sp.uns.get(SopaKeys.UNS_KEY, {}).get(SopaKeys.UNS_HAS_TRANSCRIPTS, False):
            log.warning(
                "The values in adata.X doesn't seem to be transcript count, which may create unexpected behaviors when running Tangram."
            )

        self.cell_type_key = cell_type_key
        self.reference_preprocessing = reference_preprocessing
        self.bag_size = bag_size
        self.max_obs_reference = max_obs_reference
        self.clip_percentile = clip_percentile
        self.density_prior = density_prior

        self.check_reference_annotation()

        if self.ad_sc.raw is not None:
            log.warning("The reference adata_sc has a raw attribute. Deleting it to use adata_sc.X instead.")
            del self.ad_sc.raw

        import torch

        self.device = "cuda:0" if torch.cuda.is_available() else "cpu"
        log.info(f"Using device: {self.device}")

    def _preprocess(self, adata: AnnData):
        if self.reference_preprocessing == "log1p":
            log.info("Performing normalize_total and log1p on the spatial adata object")
            sc.pp.normalize_total(adata)
            sc.pp.log1p(adata)
        elif self.reference_preprocessing == "normalized":
            log.info("Performing normalize_total on the spatial adata object")
            sc.pp.normalize_total(adata)
        else:
            log.info(
                "Not applying pre-processing on the spatial adata object (assuming it was pre-processed the same way as the reference)"
            )

    def _level_suffix(self, level: int) -> str:
        return "" if level == 0 else f"_level{level}"

    def level_obs_key(self, level: int) -> str:
        return f"{self.cell_type_key}{self._level_suffix(level)}"

    def probs_key(self, level: int) -> str:
        return f"tangram_pred{self._level_suffix(level)}"

    @property
    def levels(self):
        i = 0
        while self.level_obs_key(i) in self.ad_sc.obs.columns:
            i += 1
        return i

    def split_indices(self, indices: pd.Series, n_splits: int) -> list[np.ndarray]:
        indices = indices.values.copy()
        np.random.shuffle(indices)
        return np.array_split(indices, n_splits)

    def init_obsm(self, level: int):
        self.ad_sp.obsm[self.probs_key(level)] = pd.DataFrame(
            0.0,
            index=self.ad_sp.obs_names,
            columns=self.ad_sc.obs[self.level_obs_key(level)].unique(),
        )

    def get_hard_labels(self, df: pd.DataFrame) -> pd.Series:
        df = df.clip(df.quantile(1 - self.clip_percentile), df.quantile(self.clip_percentile), axis=1)
        df = (df - df.min()) / (df.max() - df.min())
        return df.idxmax(1)

    def pp_adata(self, ad_sp_: AnnData, ad_sc_: AnnData, split: np.ndarray) -> AnnData:
        "Copied and updated from Tangram pp_adatas()"

        ad_sp_split = ad_sp_[split].copy()
        self._preprocess(ad_sp_split)
        sc.pp.filter_genes(ad_sp_split, min_cells=1)

        # Calculate uniform density prior as 1/number_of_cells
        ad_sp_split.obs["uniform_density"] = np.ones(ad_sp_split.X.shape[0]) / ad_sp_split.X.shape[0]

        # Calculate rna_count_based density prior as % of rna molecule count
        rna_count_per_cell = np.array(ad_sp_split.X.sum(axis=1)).squeeze()
        ad_sp_split.obs["rna_count_based_density"] = rna_count_per_cell / np.sum(rna_count_per_cell)

        ad_sp_split.var["counts"] = np.array((ad_sp_split.X > 0).sum(0)).flatten()
        ad_sc_.var["counts"] = np.array((ad_sc_.X > 0).sum(0)).flatten()

        sp_lower_to_name = {gene.lower(): gene for gene in ad_sp_split.var_names}
        ad_sc_.var_names = [sp_lower_to_name.get(gene.lower(), gene) for gene in ad_sc_.var_names]

        log.info(
            f"Genes with zero counts: {(ad_sp_split.var['counts'] == 0).sum()} spatial, {(ad_sc_.var['counts'] == 0).sum()} ref"
        )

        selection = list(
            set(ad_sp_split.var_names[ad_sp_split.var.counts > 0]) & set(ad_sc_.var_names[ad_sc_.var.counts > 0])
        )

        assert len(selection), (
            "No gene in common between the reference and the spatial adata object. Have you run transcript aggregation?"
        )
        log.info(f"Keeping {len(selection)} shared genes")

        for ad_ in [ad_sp_split, ad_sc_]:
            for uns_key in ["training_genes", "overlap_genes"]:
                ad_.uns[uns_key] = selection

        return ad_sp_split

    def check_reference_annotation(self):
        assert self.cell_type_key in self.ad_sc.obs, (
            f"Cell-type key {self.cell_type_key} must be in the reference observations (adata_sc.obs)"
        )

        for level in range(1, self.levels):
            ct_key = self.level_obs_key(level)
            parent_key = self.level_obs_key(level - 1)

            for ct in self.ad_sc.obs[ct_key].unique():
                unique_parents = self.ad_sc.obs.loc[self.ad_sc.obs[ct_key] == ct, parent_key].unique()
                if len(unique_parents) != 1:
                    log.warning(
                        f"Cell-type {ct} in level {level} should have only one parent, found {list(unique_parents)}"
                    )

    def run(self):
        for level in range(self.levels):
            log.info(f"Running on level {level}")

            self.init_obsm(level)
            self.ad_sp.obs[self.level_obs_key(level)] = np.nan

            if level == 0:
                self.run_group()
                continue

            previous_key = self.level_obs_key(level - 1)
            obs_key = self.level_obs_key(level)
            self.ad_sp.obs[obs_key] = self.ad_sp.obs[previous_key]
            groups = self.ad_sc.obs.groupby(previous_key)
            for ct in groups.groups:
                group: pd.DataFrame = groups.get_group(ct)
                indices_sp = self.ad_sp.obs_names[self.ad_sp.obs[previous_key] == ct]

                sub_cts = group[obs_key].unique()
                if len(sub_cts) == 1:
                    self.ad_sp.obsm[self.probs_key(level)].loc[indices_sp, sub_cts[0]] = 1
                    continue

                log.info(f"[Cell type {ct}]")
                self.run_group(level, indices_sp, group.index)

        log.info("Finished running Tangram")

        if SopaKeys.UNS_KEY not in self.ad_sp.uns:
            self.ad_sp.uns[SopaKeys.UNS_KEY] = {}

        self.ad_sp.uns[SopaKeys.UNS_KEY][SopaKeys.UNS_CELL_TYPES] = [
            self.level_obs_key(level) for level in range(self.levels)
        ]

    def run_group(self, level: int = 0, indices_sp=None, indices_sc=None):
        try:
            import tangram as tg
        except ImportError:
            raise ImportError("To use tangram, you need to install it via `pip install tangram-sc`.")

        if indices_sp is not None and len(indices_sp) == 0:
            log.warning("No cell annotated in the upper level...")
            return

        indices_sp = self.ad_sp.obs_names if indices_sp is None else indices_sp
        ad_sp_ = self.ad_sp[indices_sp].copy()

        indices_sc = self.ad_sc.obs_names if indices_sc is None else indices_sc
        ad_sc_ = self.ad_sc[indices_sc].copy()

        if ad_sc_.n_obs >= self.max_obs_reference:
            log.info(f"Subsampling reference to {self.max_obs_reference} cells...")
            sc.pp.subsample(ad_sc_, n_obs=self.max_obs_reference)

        log.info(f"(n_obs_spatial={ad_sp_.n_obs}, n_obs_ref={ad_sc_.n_obs})")

        if not self.can_run(ad_sp_, ad_sc_):
            log.info("No annotations at this level")
            return

        n_splits = math.ceil(ad_sp_.n_obs / self.bag_size)
        for i, split in enumerate(self.split_indices(indices_sp, n_splits)):
            log.info(f"--- Split {i + 1} / {n_splits} ---")

            ad_sp_split = self.pp_adata(ad_sp_, ad_sc_, split)

            ad_map = tg.map_cells_to_space(ad_sc_, ad_sp_split, device=self.device, density_prior=self.density_prior)

            tg.project_cell_annotations(ad_map, ad_sp_split, annotation=self.level_obs_key(level))

            res = ad_sp_split.obsm["tangram_ct_pred"]
            self.ad_sp.obsm[self.probs_key(level)].loc[split, res.columns] = res

        df_group = self.ad_sp.obsm[self.probs_key(level)].loc[indices_sp]
        self.ad_sp.obs.loc[indices_sp, self.level_obs_key(level)] = self.get_hard_labels(df_group)

    def can_run(self, ad_sp_: AnnData, ad_sc_: AnnData, min_obs: int = 10) -> bool:
        if ad_sp_.n_obs < min_obs:
            log.info(f"Found only {ad_sp_.n_obs} spatial cells.")
            return False

        if ad_sc_.n_obs < min_obs:
            log.info(f"Found only {ad_sc_.n_obs} reference cells.")
            return False

        return True



from functools import partial

import dask.dataframe as dd
import geopandas as gpd
import pandas as pd
from spatialdata import SpatialData

from ..constants import SopaAttrs
from ..utils import get_boundaries, get_spatial_element, to_intrinsic


def sjoin(
    sdata: SpatialData,
    left_element: str | gpd.GeoDataFrame,
    right_element: str | gpd.GeoDataFrame,
    how: str = "left",
    target_coordinate_system: str | None = None,
    **kwargs: int,
) -> gpd.GeoDataFrame:
    """Spatial join of two `shapes` GeoDataFrames, as in [geopandas.sjoin](https://geopandas.org/en/stable/docs/reference/api/geopandas.sjoin.html).

    Shapes are automatically aligned on the same coordinate system (which can be chosen using the `target_coordinate_system` argument).

    Args:
        sdata: A `SpatialData` object
        left_element: The name of a GeoDataFrame in `sdata`, or the GeoDataFrame itself
        right_element: The name of a GeoDataFrame in `sdata`, or the GeoDataFrame itself
        how: The GeoPandas type of join. By default, left geometries are retained.
        target_coordinate_system: The name of the coordinate system on which the shapes will be transformed. By default, uses the intrinsic coordinate system of the `left_element`.
        **kwargs: Kwargs provided to the [geopandas.sjoin](https://geopandas.org/en/stable/docs/reference/api/geopandas.sjoin.html) function

    Returns:
        The joined `GeoDataFrame`
    """
    if isinstance(left_element, str):
        left_element = sdata[left_element]
    if isinstance(right_element, str):
        right_element = sdata[right_element]

    if target_coordinate_system is None:
        right_element = to_intrinsic(sdata, right_element, left_element)
    else:
        left_element = sdata.transform_element_to_coordinate_system(left_element, target_coordinate_system)
        right_element = sdata.transform_element_to_coordinate_system(right_element, target_coordinate_system)

    return gpd.sjoin(left_element, right_element, how=how, **kwargs)


def _get_cell_id(gdf: gpd.GeoDataFrame, partition: pd.DataFrame, unassigned_value: int | None = 0) -> pd.Series:
    points_gdf = gpd.GeoDataFrame(partition, geometry=gpd.points_from_xy(partition["x"], partition["y"]))
    gdf.index.name = "index_right"  # to reuse the index name later
    spatial_join = points_gdf.sjoin(gdf, how="left")
    spatial_join = spatial_join[~spatial_join.index.duplicated(keep="first")]

    cell_ids = spatial_join["index_right"]

    if unassigned_value is not None:
        cell_ids = (cell_ids.fillna(-1) + 1 + unassigned_value).astype(int)
    else:
        cell_ids = cell_ids.astype("Int64")  # integers with NaNs

    return cell_ids


def assign_transcript_to_cell(
    sdata: SpatialData,
    points_key: str | None = None,
    shapes_key: str | None = None,
    key_added: str = "cell_index",
    unassigned_value: int | None = None,
):
    """Assign each transcript to a cell based on the cell boundaries. It updates the transcript dataframe by adding a new column.

    Args:
        sdata: A `SpatialData` object
        points_key: Key of the spatialdata object containing the transcript dataframe.
        shapes_key: Key of the spatialdata object containing the cell boundaries.
        key_added: Key that will be added to the transcript dataframe containing the cell ID
        unassigned_value: If `None`, transcripts that are not inside any cell will be assigned to NaN. If an integer, this value will be used as the unassigned value.
    """
    df = get_spatial_element(sdata.points, points_key or sdata.attrs.get(SopaAttrs.TRANSCRIPTS))
    geo_df = get_boundaries(sdata, key=shapes_key)

    geo_df = to_intrinsic(sdata, geo_df, df)
    geo_df = geo_df.reset_index(drop=True)

    get_cell_id = partial(_get_cell_id, geo_df, unassigned_value=unassigned_value)

    if isinstance(df, dd.DataFrame):
        df[key_added] = df.map_partitions(get_cell_id)
    else:
        raise TypeError(f"Invalid dataframe type: {type(df)}")



import numpy as np


def dummy_method(**method_kwargs):
    """A method builder builder (i.e. it returns a segmentation function).
    Kwargs can be provided and used in the below function"""

    def segmentation_function(image: np.ndarray) -> np.ndarray:
        """A dummy example of a custom segmentation method
        that creates one cell (with a padding of 10 pixels).

        Args:
            image: An image of shape `(C, Y, X)`

        Returns:
            A mask of shape `(Y, X)` containing one cell
        """
        mask = np.zeros(image.shape[1:], dtype=int)

        # one cell, corresponding to value 1
        mask[10:-10, 10:-10] = 1  # squared shaped

        return mask

    return segmentation_function



import typer

from .utils import SDATA_HELPER

app_patchify = typer.Typer()


@app_patchify.command()
def image(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    patch_width_pixel: float = typer.Option(5000, help="Width (and height) of each patch in pixels"),
    patch_overlap_pixel: float = typer.Option(
        100,
        help="Number of overlapping pixels between the patches. We advise to choose approximately twice the diameter of a cell",
    ),
):
    """Prepare patches for staining-based segmentation (including Cellpose)"""
    import sopa
    from sopa.constants import SopaFiles, SopaKeys
    from sopa.io.standardize import read_zarr_standardized

    sdata = read_zarr_standardized(sdata_path)

    sopa.make_image_patches(sdata, patch_width_pixel, patch_overlap_pixel)

    n_patches = len(sdata[SopaKeys.PATCHES])
    _save_cache(sdata_path, SopaFiles.PATCHES_FILE_IMAGE, n_patches)


@app_patchify.command()
def transcripts(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    patch_width_microns: float = typer.Option(5000, help="Width (and height) of each patch in microns"),
    patch_overlap_microns: float = typer.Option(
        100,
        help="Number of overlapping microns between the patches. We advise to choose approximately twice the diameter of a cell",
    ),
    unassigned_value: str = typer.Option(
        None,
        help="If --prior-shapes-key is provided, this is the value given to transcripts that are not inside any cell (if it's already 0, don't provide this argument)",
    ),
    prior_shapes_key: str = typer.Option(
        None,
        help="Name of the boundaries element to use as a segmentation prior",
    ),
    write_cells_centroids: bool = typer.Option(
        False,
        help="Whether to also write the centroids of the cells (must be True for ComSeg)",
    ),
):
    """Prepare patches for transcript-based segmentation (e.g., for Proseg or Baysor)"""

    import sopa
    from sopa.constants import SopaFiles, SopaKeys
    from sopa.io.standardize import read_zarr_standardized

    sdata = read_zarr_standardized(sdata_path)

    if isinstance(unassigned_value, str) and unassigned_value.isdigit():
        unassigned_value = int(unassigned_value)

    sopa.make_transcript_patches(
        sdata,
        patch_width_microns,
        patch_overlap_microns,
        unassigned_value=unassigned_value,
        prior_shapes_key=prior_shapes_key,
        write_cells_centroids=write_cells_centroids,
    )

    valid_indices = list(sdata[SopaKeys.TRANSCRIPTS_PATCHES].index)
    _save_cache(sdata_path, SopaFiles.PATCHES_FILE_TRANSCRIPTS, "\n".join(map(str, valid_indices)))


def _save_cache(sdata_path: str, filename: str, content: str):
    from pathlib import Path

    from sopa.constants import SopaFiles

    cache_file = Path(sdata_path) / SopaFiles.SOPA_CACHE_DIR / filename
    cache_file.parent.mkdir(parents=True, exist_ok=True)

    with open(cache_file, "w", newline="\n") as f:
        f.write(f"{content}\n")



import logging

import numpy as np
import shapely
import shapely.affinity
from shapely.geometry import MultiPolygon, Polygon
from skimage.draw import polygon

log = logging.getLogger(__name__)


def rasterize(cell: Polygon | MultiPolygon, shape: tuple[int, int], xy_min: tuple[int, int] = (0, 0)) -> np.ndarray:
    """Transform a cell polygon into a numpy array with value 1 where the polygon touches a pixel, else 0.

    Args:
        cell: Cell polygon to rasterize.
        shape: Image shape as a tuple (y, x).
        xy_min: Tuple containing the origin of the image [x0, y0].

    Returns:
        The mask array.
    """
    xmin, ymin, xmax, ymax = [xy_min[0], xy_min[1], xy_min[0] + shape[1], xy_min[1] + shape[0]]

    cell_translated = shapely.affinity.translate(cell, -xmin, -ymin)
    geoms = cell_translated.geoms if isinstance(cell_translated, MultiPolygon) else [cell_translated]

    shape = (ymax - ymin, xmax - xmin)

    rasterized_image = np.zeros(shape, dtype=np.int8)

    for geom in geoms:
        x, y = geom.exterior.coords.xy
        rr, cc = polygon(y, x, shape)
        rasterized_image[rr, cc] = 1

    return rasterized_image



import logging
import warnings

import matplotlib.pyplot as plt
import numpy as np
import scanpy as sc
import seaborn as sns
from anndata import AnnData
from spatialdata import SpatialData

from ...constants import LOW_AVERAGE_COUNT, SopaKeys
from ...utils import get_boundaries, get_intensities, get_spatial_image
from .engine import CodeBlock, Columns, Image, Message, Paragraph, Root, Section, SubSection

log = logging.getLogger(__name__)


def write_report(path: str, sdata: SpatialData, table_key: str = SopaKeys.TABLE):
    """Create a HTML report (or web report) after running Sopa.

    Note:
        This report is automatically generated based on a custom python-to-html engine

    Args:
        path: Path to the `.html` report that has to be created
        sdata: A `SpatialData` object, after running Sopa
        table_key: Key of the table in the `SpatialData` object to be used for the report
    """
    with warnings.catch_warnings():
        warnings.simplefilter(action="ignore", category=FutureWarning)

        sections = SectionBuilder(sdata, table_key).compute_sections()

        log.info(f"Writing report to {path}")
        Root(sections).write(path)


def _kdeplot_vmax_quantile(values: np.ndarray, quantile: float = 0.95):
    threshold = np.quantile(values, quantile)
    sns.kdeplot(values)
    plt.xlim(0, threshold)


class SectionBuilder:
    SECTION_NAMES = [
        "general_section",
        "cell_section",
        "channel_section",
        "transcripts_section",
        "representation_section",
    ]

    def __init__(self, sdata: SpatialData, table_key: str):
        self.sdata = sdata
        self.table_key = table_key

        if table_key not in self.sdata.tables:
            log.warning(f"Table key '{table_key}' not found in the SpatialData object")

        self.adata = None
        if table_key in self.sdata.tables:
            self.adata: AnnData = self.sdata.tables[table_key].copy()

    def _table_has(self, key, default=False):
        if SopaKeys.UNS_KEY not in self.adata.uns:
            return default
        return self.adata.uns[SopaKeys.UNS_KEY].get(key, default)

    def general_section(self):
        return Section(
            "General",
            [
                SubSection(
                    "SpatialData information",
                    [
                        Paragraph(
                            "Sopa is using <a href='https://spatialdata.scverse.org/en/latest/'>SpatialData</a> under the hood. This is how the object looks like:"
                        ),
                        CodeBlock(str(self.sdata)),
                    ],
                )
            ],
        )

    def cell_section(self):
        shapes_key, _ = get_boundaries(self.sdata, return_key=True, table_key=self.table_key)

        fig = plt.figure()
        _kdeplot_vmax_quantile(self.adata.obs[SopaKeys.AREA_OBS])
        plt.xlabel("Area (coordinate_system_unit ^ 2)")

        return Section(
            "Cells",
            [
                SubSection(
                    "Number",
                    Paragraph(f"Number of cells:<br>{Message(f'{self.adata.n_obs} cells')}"),
                ),
                SubSection(
                    "Areas",
                    [
                        Paragraph(
                            f"The cells areas are obtained based on the '{shapes_key}' boundaries (and its intrinsic coordinate system)."
                        ),
                        Columns([Image(fig)]),
                    ],
                ),
            ],
        )

    def channel_section(self):
        image = get_spatial_image(self.sdata)

        subsections = [
            SubSection(
                "Names",
                Paragraph(f"Channels names:<br>{Message(', '.join(map(str, list(image.coords['c'].values))))}"),
            )
        ]

        if self._table_has(SopaKeys.UNS_HAS_INTENSITIES):
            fig = plt.figure()

            df_intensities = get_intensities(self.sdata, self.table_key)
            threshold = np.quantile(df_intensities.values.ravel(), 0.95)

            for channel, intensities in df_intensities.items():
                sns.kdeplot(intensities, label=channel)
            plt.xlim(0, threshold)
            plt.xlabel("Intensity")
            plt.ylabel("Distribution density")

            subsections.append(
                SubSection(
                    "Per-cell intensity",
                    [
                        Paragraph(
                            "Each channel is averaged over each cell. We display the distribution of channel intensities over all cells:"
                        ),
                        Columns([Image(fig)]),
                    ],
                )
            )

        return Section("Channels", subsections)

    def transcripts_section(self):
        if not self._table_has(SopaKeys.UNS_HAS_TRANSCRIPTS):
            return None

        counts = self.adata.layers.get("counts", self.adata.X)

        mean_transcript_count = _to_array(counts.mean(0))
        low_average = mean_transcript_count < LOW_AVERAGE_COUNT

        QC_subsubsections = []
        if low_average.sum():
            QC_subsubsections.append(
                Paragraph(
                    f"{low_average.sum()} genes have a low count (less than {LOW_AVERAGE_COUNT} per cell, on average):"
                )
            )
            genes = self.adata.var_names[low_average].tolist()
            if len(genes) > 50:
                genes = [*genes[:50], f"and {len(genes) - 50} others..."]
            QC_subsubsections.append(Message(", ".join(genes), color="danger"))

        fig1 = plt.figure()
        _kdeplot_vmax_quantile(mean_transcript_count)
        plt.xlabel("Count per transcript (average / cells)")

        counts = _to_array(counts.sum(1))
        self.adata.obs["transcript_counts"] = counts

        fig2 = plt.figure()
        _kdeplot_vmax_quantile(counts)
        plt.xlabel("Transcript count per cell")

        QC_subsubsections.append(Columns([Image(fig1), Image(fig2)]))

        x, y = self.adata.obsm["spatial"].T
        spot_size = np.sqrt((x.max() - x.min()) * (y.max() - y.min()) / self.adata.n_obs)  # visually pleasing
        fig3 = sc.pl.spatial(
            self.adata, color="transcript_counts", spot_size=spot_size, return_fig=True, show=False, vmin=0, vmax="p95"
        )

        return Section(
            "Transcripts",
            [
                SubSection("Quality controls", QC_subsubsections),
                SubSection("Spatial distribution", Columns([Image(fig3)])),
            ],
        )

    def representation_section(self, max_obs: int = 100_000):
        if "X_umap" in self.adata.obsm:
            log.info("UMAP already computed")
            return self._umap_figure()

        if self._table_has(SopaKeys.UNS_HAS_TRANSCRIPTS):
            sc.pp.normalize_total(self.adata)
            sc.pp.log1p(self.adata)

        if self.adata.n_obs > max_obs:
            sc.pp.subsample(self.adata, n_obs=max_obs)

        log.info(f"Computing UMAP on {self.adata.n_obs} cells")

        sc.pp.pca(self.adata)
        sc.pp.neighbors(self.adata)
        sc.tl.umap(self.adata)

        return self._umap_figure()

    def _umap_figure(self):
        default_color = "leiden" if "leiden" in self.adata.obs else None
        color = self._table_has(SopaKeys.UNS_CELL_TYPES, default_color)

        sc.pl.umap(self.adata, color=color, show=False)

        return Section(
            "Representation",
            [
                SubSection("UMAP", Columns([Image(plt.gcf(), pretty_legend=False)])),
            ],
        )

    def compute_sections(self) -> list[Section]:
        sections = []

        for name in self.SECTION_NAMES:
            try:
                log.info(f"Writing {name}")
                section = getattr(self, name)()
                sections.append(section)
            except Exception as e:
                log.warning(f"Section {name} failed with error {e}")

        return [section for section in sections if section is not None]


def _to_array(value: np.ndarray | np.matrix) -> np.ndarray:
    if isinstance(value, np.matrix):
        return value.A1
    if isinstance(value, np.ndarray):
        return value
    raise TypeError(f"Unsupported type: {type(value)}")



import json
import logging
import shutil
from pathlib import Path

import anndata
import geopandas as gpd
import numpy as np
import pandas as pd
from anndata import AnnData
from shapely.geometry import Polygon, shape
from spatialdata import SpatialData
from spatialdata.models import ShapesModel, TableModel
from tqdm import tqdm

from ... import shapes
from ...aggregation import count_transcripts
from ...constants import SopaKeys
from ...utils import add_spatial_element, copy_transformations, get_feature_key, get_transcripts_patches_dirs
from .. import solve_conflicts

log = logging.getLogger(__name__)


def resolve(
    sdata: SpatialData,
    patches_dirs: list[str | Path],
    gene_column: str | None = None,
    min_area: float = 0,
    key_added: str = SopaKeys.BAYSOR_BOUNDARIES,
):
    """Concatenate all the per-patch segmentation runs and resolve the conflicts. Resulting cells boundaries are saved in the `SpatialData` object.

    Args:
        sdata: A `SpatialData` object
        gene_column: Column of the transcript dataframe containing the genes names
        patches_dirs: Optional list of subdirectories inside `temp_dir` to be read. By default, read all.
        min_area: Minimum area (in microns^2) for a cell to be kept
        key_added: Name of the spatial element that will be added, containing the cell boundaries
    """
    if min_area > 0:
        log.info(f"Cells whose area is less than {min_area} microns^2 will be removed")

    patches_cells, adatas = _read_all_segmented_patches(patches_dirs, min_area)
    geo_df, cells_indices, new_ids = _resolve_patches(patches_cells, adatas)

    points_key = sdata[SopaKeys.TRANSCRIPTS_PATCHES][SopaKeys.POINTS_KEY].iloc[0]
    points = sdata[points_key]
    transformations = copy_transformations(points)

    geo_df = ShapesModel.parse(geo_df, transformations=transformations)

    table_conflicts = []
    if len(new_ids):
        new_cells = geo_df.geometry[cells_indices == -1]
        geo_df_new = gpd.GeoDataFrame({"geometry": new_cells})
        geo_df_new = ShapesModel.parse(geo_df_new, transformations=transformations)

        log.info("Aggregating transcripts on merged cells")
        gene_column = gene_column or get_feature_key(sdata, raise_error=True)
        table_conflicts = count_transcripts(sdata, gene_column, geo_df=geo_df_new, points_key=points_key)
        table_conflicts.obs_names = new_ids
        table_conflicts = [table_conflicts]

    valid_ids = set(geo_df.index)
    table = anndata.concat(
        [adata[list(valid_ids & set(adata.obs_names))] for adata in adatas] + table_conflicts,
        join="outer",
    )
    table.obs.dropna(axis="columns", inplace=True)

    geo_df = geo_df.loc[table.obs_names]

    table.obsm["spatial"] = np.array([[centroid.x, centroid.y] for centroid in geo_df.centroid])
    table.obs[SopaKeys.REGION_KEY] = pd.Series(key_added, index=table.obs_names, dtype="category")
    table.obs[SopaKeys.INSTANCE_KEY] = geo_df.index

    table = TableModel.parse(
        table,
        region_key=SopaKeys.REGION_KEY,
        region=key_added,
        instance_key=SopaKeys.INSTANCE_KEY,
    )

    add_spatial_element(sdata, key_added, geo_df)
    add_spatial_element(sdata, SopaKeys.TABLE, table)

    log.info(f"Added sdata.tables['{SopaKeys.TABLE}'], and {len(geo_df)} cell boundaries to sdata['{key_added}']")


def _read_one_segmented_patch(
    directory: str, min_area: float = 0, min_vertices: int = 4
) -> tuple[list[Polygon], AnnData]:
    directory: Path = Path(directory)
    id_as_string, polygon_file = _find_polygon_file(directory)

    loom_file = directory / "segmentation_counts.loom"
    if loom_file.exists():
        adata = anndata.io.read_loom(directory / "segmentation_counts.loom", obs_names="Name", var_names="Name")
    else:
        adata = anndata.io.read_h5ad(directory / "segmentation_counts.h5ad")

    adata.obs.rename(columns={"area": SopaKeys.ORIGINAL_AREA_OBS}, inplace=True)

    cells_ids = pd.Series(
        adata.obs_names if id_as_string else adata.obs["CellID"].astype(int),
        index=adata.obs_names,
    )
    del adata.obs["CellID"]

    with open(polygon_file) as f:
        polygons_dict = json.load(f)
        polygons_dict = {c["cell"]: c for c in polygons_dict["geometries"]}

    def _keep_cell(ID: str | int):
        if ID not in polygons_dict:
            return False
        return len(polygons_dict[ID]["coordinates"][0]) >= min_vertices

    cells_ids = cells_ids[cells_ids.map(_keep_cell)]

    geo_df = gpd.GeoDataFrame(index=cells_ids.index, geometry=[shape(polygons_dict[ID]) for ID in cells_ids])
    geo_df = shapes.to_valid_polygons(geo_df)

    ratio_filtered = (geo_df.area <= min_area).mean()
    if ratio_filtered > 0.2:
        log.warning(f"{ratio_filtered:.2%} of cells will be filtered due to {min_area=}")

    geo_df = geo_df[geo_df.area > min_area]

    return geo_df.geometry.values, adata[geo_df.index].copy()


def _find_polygon_file(directory: Path) -> tuple[bool, Path]:
    old_baysor_path = directory / "segmentation_polygons.json"
    if old_baysor_path.exists():
        return False, old_baysor_path
    new_baysor_path = directory / "segmentation_polygons_2d.json"
    assert new_baysor_path.exists(), f"Could not find the segmentation polygons file in {directory}"
    return True, new_baysor_path


def _read_all_segmented_patches(
    patches_dirs: list[str],
    min_area: float = 0,
) -> tuple[list[list[Polygon]], list[AnnData]]:
    outs = [
        _read_one_segmented_patch(path, min_area)
        for path in tqdm(patches_dirs, desc="Reading transcript-segmentation outputs")
    ]

    patches_cells, adatas = zip(*outs)

    return patches_cells, adatas


def _resolve_patches(
    patches_cells: list[list[Polygon]], adatas: list[AnnData]
) -> tuple[gpd.GeoDataFrame, np.ndarray, np.ndarray]:
    """Resolve the segmentation conflits on the patches overlaps.

    Args:
        patches_cells: List of polygons segmented on each patch
        adatas: List of AnnData objects corresponding to each patch

    Returns:
        The new GeoDataFrame, the new cells indices (-1 for merged cells), and the ids of the merged cells.
    """
    patch_ids = [adata.obs_names for adata in adatas]

    patch_indices = np.arange(len(patches_cells)).repeat([len(cells) for cells in patches_cells])
    cells = [cell for cells in patches_cells for cell in cells]
    segmentation_ids = np.array([cell_id for ids in patch_ids for cell_id in ids])

    cells_resolved, cells_indices = solve_conflicts(cells, patch_indices=patch_indices, return_indices=True)

    existing_ids = segmentation_ids[cells_indices[cells_indices >= 0]]
    new_ids = np.char.add("merged_cell_", np.arange((cells_indices == -1).sum()).astype(str))
    cells_resolved.index = np.concatenate([existing_ids, new_ids])

    return cells_resolved, cells_indices, new_ids


def _check_transcript_patches(sdata: SpatialData, with_prior: bool = False):
    assert SopaKeys.TRANSCRIPTS_PATCHES in sdata.shapes, (
        "Transcript patches not found in the SpatialData object. Run `sopa.make_transcript_patches(...)` first."
    )

    directories: list[Path] = get_transcripts_patches_dirs(sdata)

    assert all(directory.exists() for directory in directories), (
        "Some patch directories are missing. "
        "This can happen if you already finished a segmentation which deleted the cache. "
        "You must re-run `sopa.make_transcript_patches`. "
        "Next time, use `delete_cache=False` during the segmentation to keep the cache."
    )

    if with_prior:
        assert SopaKeys.PRIOR_SHAPES_KEY in sdata[SopaKeys.TRANSCRIPTS_PATCHES].columns, (
            "You need to run `sopa.make_transcript_patches` with a `prior_shapes_key`. "
            "You can provide `prior_shapes_key='auto'` if your technology has a prior segmentation, or `prior_shapes_key='cellpose_boundaries'` if you ran cellpose segmentation first."
        )


def _get_executable_path(name: str, default_dir: str) -> Path | str:
    if shutil.which(name) is not None:
        return name

    default_path = Path.home() / default_dir / "bin" / name
    if default_path.exists():
        return default_path

    bin_path = Path.home() / ".local" / "bin" / name
    raise FileNotFoundError(
        f"Please install {name} and ensure that either `{default_path}` executes {name}, or that `{name}` is an existing command (add it to your PATH, or create a symlink at {bin_path})."
    )



import ast
from collections.abc import Iterable
from typing import Any

import typer

from .utils import SDATA_HELPER, _log_whether_to_resolve

app_segmentation = typer.Typer()


@app_segmentation.command()
def cellpose(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    diameter: int = typer.Option(help="Cellpose diameter parameter"),
    channels: list[str] = typer.Option(
        help="Names of the channels used for Cellpose. If one channel, then provide just a nucleus channel. If two channels, this is the cytoplasm and then the nucleus channel"
    ),
    flow_threshold: float = typer.Option(2, help="Cellpose `flow_threshold` parameter"),
    cellprob_threshold: float = typer.Option(-6, help="Cellpose `cellprob_threshold` parameter"),
    model_type: str = typer.Option("cyto3", help="Name of the cellpose model"),
    pretrained_model: str = typer.Option(None, help="Path to the pretrained model to be loaded"),
    gpu: bool = typer.Option(False, help="Whether to use the GPU (recommended for cellpose>=4.0.0)"),
    min_area: int = typer.Option(0, help="Minimum area (in pixels^2) for a cell to be considered as valid"),
    clip_limit: float = typer.Option(
        0.2,
        help="Parameter for skimage.exposure.equalize_adapthist (applied before running cellpose)",
    ),
    clahe_kernel_size: int = typer.Option(
        None,
        help="Parameter for skimage.exposure.equalize_adapthist (applied before running cellpose)",
    ),
    gaussian_sigma: float = typer.Option(
        1, help="Parameter for scipy gaussian_filter (applied before running cellpose)"
    ),
    patch_index: int = typer.Option(
        default=None,
        help="Index of the patch on which cellpose should be run. NB: the number of patches is `len(sdata['image_patches'])`",
    ),
    cache_dir_name: str = typer.Option(
        default=None,
        help="Name of the temporary cellpose directory inside which we will store each individual patch segmentation. By default, uses the `cellpose_boundaries` directory",
    ),
    method_kwargs: str = typer.Option(
        {},
        callback=ast.literal_eval,
        help="Kwargs for the cellpose method builder. This should be a dictionnary, in inline string format.",
    ),
):
    """Perform cellpose segmentation. This can be done on all patches directly, or on one individual patch.

    Usage:

        - [On one patch] Use this mode to run patches in parallel. Provide `--patch-index` to run one patch, and execute all patches in a parallel manner (you need to define your own parallelization, else, use the Snakemake pipeline).

        - [On all patches at once] For small images, you can run the segmentation method sequentially (`--patch-index` is not needed)
    """
    from sopa.constants import SopaKeys

    channels = channels if isinstance(channels, list) else [channels]

    _run_staining_segmentation(
        sdata_path,
        "cellpose_patch",
        SopaKeys.CELLPOSE_BOUNDARIES,
        channels,
        min_area,
        clip_limit,
        clahe_kernel_size,
        gaussian_sigma,
        patch_index,
        cache_dir_name,
        diameter=diameter,
        flow_threshold=flow_threshold,
        cellprob_threshold=cellprob_threshold,
        model_type=model_type,
        pretrained_model=pretrained_model,
        gpu=gpu,
        **method_kwargs,
    )


@app_segmentation.command()
def stardist(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    model_type: str = typer.Option("2D_versatile_he", help="Name of the stardist model"),
    prob_thresh: float = typer.Option(0.2, help="Stardist `prob_thresh` parameter."),
    nms_thresh: float = typer.Option(0.6, help="Stardist `nms_thresh` parameter."),
    channels: list[str] = typer.Option(None, help="Names of the channels used for segmentation."),
    min_area: int = typer.Option(0, help="Minimum area (in pixels^2) for a cell to be considered as valid"),
    clip_limit: float = typer.Option(
        0.2,
        help="Parameter for skimage.exposure.equalize_adapthist (applied before running the segmentation method)",
    ),
    clahe_kernel_size: int = typer.Option(
        None,
        help="Parameter for skimage.exposure.equalize_adapthist (applied before running stardist)",
    ),
    gaussian_sigma: float = typer.Option(
        1,
        help="Parameter for scipy gaussian_filter (applied before running the segmentation method)",
    ),
    patch_index: int = typer.Option(
        default=None,
        help="Index of the patch on which the segmentation method should be run. NB: the number of patches is `len(sdata['image_patches'])`",
    ),
    cache_dir_name: str = typer.Option(
        default=None,
        help="Name of the temporary the segmentation method directory inside which we will store each individual patch segmentation. By default, uses the `stardist_boundaries` directory",
    ),
    method_kwargs: str = typer.Option(
        {},
        callback=ast.literal_eval,
        help="Kwargs for the method. This should be a dictionnary, in inline string format.",
    ),
):
    """Perform Stardist segmentation. This can be done on all patches directly, or on one individual patch."""
    from sopa.constants import SopaKeys

    _run_staining_segmentation(
        sdata_path,
        "stardist_patch",
        SopaKeys.STARDIST_BOUNDARIES,
        channels,
        min_area,
        clip_limit,
        clahe_kernel_size,
        gaussian_sigma,
        patch_index,
        cache_dir_name,
        prob_thresh=prob_thresh,
        nms_thresh=nms_thresh,
        model_type=model_type,
        **method_kwargs,
    )


@app_segmentation.command()
def generic_staining(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    method_name: str = typer.Option(
        help="Name of the segmentation method builder to use. The corresponding function (`sopa.segmentation.methods.<method_name>`) will be used, and the kwargs below will be used to instantiate the method."
    ),
    method_kwargs: str = typer.Option(
        {},
        callback=ast.literal_eval,
        help="Kwargs for the method. This should be a dictionnary, in inline string format.",
    ),
    channels: list[str] = typer.Option(None, help="Names of the channels used for segmentation."),
    min_area: int = typer.Option(0, help="Minimum area (in pixels^2) for a cell to be considered as valid"),
    clip_limit: float = typer.Option(
        0.2,
        help="Parameter for skimage.exposure.equalize_adapthist (applied before running the segmentation method)",
    ),
    clahe_kernel_size: int = typer.Option(
        None,
        help="Parameter for skimage.exposure.equalize_adapthist (applied before running segmentation)",
    ),
    gaussian_sigma: float = typer.Option(
        1,
        help="Parameter for scipy gaussian_filter (applied before running the segmentation method)",
    ),
    patch_index: int = typer.Option(
        default=None,
        help="Index of the patch on which the segmentation method should be run. NB: the number of patches is `len(sdata['image_patches'])`",
    ),
    cache_dir_name: str = typer.Option(
        default=None,
        help="Name of the temporary the segmentation method directory inside which we will store each individual patch segmentation. By default, uses the `<method_name>` directory",
    ),
):
    """Perform generic staining-based segmentation. This can be done on all patches directly, or on one individual patch.

    Usage:
        First, define a new segmentation method, and write it under `sopa.segmentation.methods`. It should correspond to a function that is a "callable builder", i.e. kwargs will be provided to this function, and it will return a callable that will be applied on patches.

        As for Cellpose, two modes ara available:

        - [On one patch] Use this mode to run patches in parallel. Provide `--patch-index` to run one patch, and execute all patches in a parallel manner (you need to define your own parallelization, else, use the Snakemake pipeline).

        - [On all patches at once] For small images, you can run the segmentation method sequentially (`--patch-index` is not needed)
    """
    from sopa.segmentation import methods

    assert hasattr(methods, method_name), (
        f"'{method_name}' is not a valid method builder under `sopa.segmentation.methods`"
    )

    _run_staining_segmentation(
        sdata_path,
        method_name,
        method_name,
        channels,
        min_area,
        clip_limit,
        clahe_kernel_size,
        gaussian_sigma,
        patch_index,
        cache_dir_name,
        **method_kwargs,
    )


def _run_staining_segmentation(
    sdata_path: str,
    method_name: str,
    key_added: str,
    channels: list[str] | None,
    min_area: float,
    clip_limit: float,
    clahe_kernel_size: int | Iterable[int] | None,
    gaussian_sigma: float,
    patch_index: int | None,
    cache_dir_name: str | None,
    **method_kwargs: Any,
):
    from sopa.io.standardize import read_zarr_standardized
    from sopa.segmentation import StainingSegmentation, custom_staining_based, methods

    from .utils import _default_boundary_dir

    sdata = read_zarr_standardized(sdata_path)

    if channels is not None:
        method_kwargs["channels"] = channels

    method = getattr(methods, method_name)(**method_kwargs)

    delete_cache = cache_dir_name is None
    if cache_dir_name is None:
        cache_dir_name = key_added

    if patch_index is None:
        custom_staining_based(
            sdata,
            method=method,
            channels=channels,
            min_area=min_area,
            clip_limit=clip_limit,
            clahe_kernel_size=clahe_kernel_size,
            gaussian_sigma=gaussian_sigma,
            key_added=key_added,
            cache_dir_name=cache_dir_name,
            delete_cache=delete_cache,
        )
        _log_whether_to_resolve(patch_index, delete_cache=delete_cache)
        return

    segmentation = StainingSegmentation(
        sdata,
        method,
        channels,
        min_area=min_area,
        clip_limit=clip_limit,
        clahe_kernel_size=clahe_kernel_size,
        gaussian_sigma=gaussian_sigma,
    )

    patch_dir = _default_boundary_dir(sdata_path, cache_dir_name)

    segmentation.write_patch_cells(patch_dir, patch_index)


@app_segmentation.command()
def comseg(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    config: str = typer.Option(
        {},
        callback=ast.literal_eval,
        help="""Comseg config. This should be a dictionnary in inline string format, wrapped inside quotes, e.g. '{"example": 2}'""",
    ),
    patch_index: int = typer.Option(
        default=None,
        help="Index of the patch on which the segmentation method should be run.",
    ),
    min_area: float = typer.Option(default=0, help="Minimum area (in micron^2) for a cell to be considered as valid"),
):
    """Perform ComSeg segmentation. This can be done on all patches directly, or on one individual patch."""
    from sopa.io.standardize import read_zarr_standardized
    from sopa.segmentation.methods import comseg

    sdata = read_zarr_standardized(sdata_path)

    comseg(sdata, config=config, min_area=min_area, patch_index=patch_index)

    _log_whether_to_resolve(patch_index)


@app_segmentation.command()
def baysor(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    config: str = typer.Option(
        {},
        callback=ast.literal_eval,
        help="""Baysor config. This should be a dictionnary in inline string format, wrapped inside single quotes, e.g. '{"example": 2}'""",
    ),
    patch_index: int = typer.Option(
        default=None,
        help="Index of the patch on which the segmentation method should be run. By default, run on all patches.",
    ),
    min_area: float = typer.Option(default=0, help="Minimum area (in micron^2) for a cell to be considered as valid"),
    scale: float = typer.Option(default=None, help="Baysor scale parameter (for config inference)"),
):
    """Perform Baysor segmentation. This can be done on all patches directly, or on one individual patch."""
    import sys
    from subprocess import CalledProcessError

    from sopa.io.standardize import read_zarr_standardized
    from sopa.segmentation.methods import baysor

    sdata = read_zarr_standardized(sdata_path)

    try:
        baysor(sdata, config=config, min_area=min_area, patch_index=patch_index, scale=scale)
    except CalledProcessError as e:
        sys.exit(e.returncode)

    _log_whether_to_resolve(patch_index)


@app_segmentation.command()
def proseg(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    command_line_suffix: str = typer.Option(
        "",
        help="String suffix to add to the proseg command line. This can be used to add extra parameters to the proseg command line.",
    ),
    infer_presets: bool = typer.Option(
        True,
        help="Whether to infer the proseg presets based on the columns of the transcripts dataframe.",
    ),
    prior_shapes_key: str | None = typer.Option(
        None,
        help="**Only for Visium HD data.** Key of `sdata` containing the prior cell boundaries. If `'auto'`, use the latest performed segmentation (e.g., stardist or the 10X Genomics segmentation).",
    ),
    bins_key: str = typer.Option(
        None,
        help="**Only for Visium HD data.** Key of `sdata` with the table corresponding to the bin-by-gene table of gene counts (e.g., for Visium HD data). Inferred by default.",
    ),
):
    """Perform Proseg segmentation. This needs to be done on a single
    patch as proseg is fast enough and doesn't require parallelization."""
    import sys
    from subprocess import CalledProcessError

    from sopa.io.standardize import read_zarr_standardized
    from sopa.segmentation.methods import proseg

    sdata = read_zarr_standardized(sdata_path)

    try:
        proseg(
            sdata,
            command_line_suffix=command_line_suffix,
            infer_presets=infer_presets,
            prior_shapes_key=prior_shapes_key,
            bins_key=bins_key,
        )
    except CalledProcessError as e:
        sys.exit(e.returncode)

    _log_whether_to_resolve(None)


@app_segmentation.command()
def tissue(
    sdata_path: str = typer.Argument(help=SDATA_HELPER),
    image_key: str = typer.Option(
        default=None,
        help="Name of the image key to use for tissue segmentation.",
    ),
    level: int = typer.Option(
        default=-1,
        help="Level of the multiscale image to use for tissue segmentation.",
    ),
    mode: str = typer.Option(
        default=None,
        help="Mode for the tissue segmentation: 'staining' or 'saturation' (for H&E images).",
    ),
    kwargs: str = typer.Option(
        {},
        callback=ast.literal_eval,
        help="Kwargs for `sopa.segmentation.tissue`. This should be a dictionnary, in inline string format.",
    ),
):
    """Perform tissue segmentation. This can be done only on objects with H&E staining."""
    import sopa
    from sopa.io.standardize import read_zarr_standardized

    sdata = read_zarr_standardized(sdata_path)

    sopa.segmentation.tissue(sdata, image_key=image_key, level=level, mode=mode, **kwargs)
